"""Tests for the solver module in Pastas."""

from unittest.mock import MagicMock, patch

import numpy as np
import pandas as pd
import pytest
from pandas import DataFrame, Series

import pastas as ps
from pastas.solver import BaseSolver, EmceeSolve, LeastSquares, LmfitSolve


@pytest.fixture
def mock_model() -> MagicMock:
    """Create a mock model for solver testing."""
    model = MagicMock()

    # Setup parameters
    parameters = DataFrame(
        data={
            "initial": [1.0, 2.0, 3.0],
            "pmin": [0.1, 0.2, 0.3],
            "pmax": [10.0, 20.0, 30.0],
            "vary": [True, True, False],
            "name": ["param1", "param2", "param3"],
            "stderr": [0.1, 0.2, 0.3],
        },
        index=["p1", "p2", "p3"],
    )
    model.parameters = parameters

    # Setup time series data
    dates = pd.date_range(start="2000-01-01", periods=100, freq="D")
    residuals = Series(np.random.normal(0, 1, 100), index=dates)
    noise = Series(np.random.normal(0, 0.5, 100), index=dates)

    # Mock methods
    model.residuals.return_value = residuals
    model.noise.return_value = noise
    model.noise_weights.return_value = Series(np.ones(100), index=dates)
    model.get_parameters.return_value = np.array([1.0, 2.0])
    model.simulate.return_value = Series(
        np.random.normal(5, 1, 100), index=dates, name="Simulation"
    )
    model.get_contribution.return_value = Series(
        np.random.normal(2, 0.5, 100), index=dates, name="Contribution"
    )
    model.get_block_response.return_value = Series(
        np.exp(-np.arange(50) / 10), index=np.arange(50), name="Block Response"
    )
    model.get_step_response.return_value = Series(
        1 - np.exp(-np.arange(50) / 10), index=np.arange(50), name="Step Response"
    )

    return model


# Existing integration tests with real models
def test_least_squares(ml_recharge: ps.Model) -> None:
    ml_recharge.solve(solver=ps.LeastSquares())


def test_least_squares_lm(ml_recharge: ps.Model) -> None:
    ml_recharge.solve(solver=ps.LeastSquares(), method="lm")
    assert ml_recharge.parameters.loc[ml_recharge.parameters.vary, "pmin"].isna().all()


def test_fit_constant(ml_recharge: ps.Model) -> None:
    ml_recharge.solve(fit_constant=False)


def test_no_noise(ml_recharge: ps.Model) -> None:
    ml_recharge.del_noisemodel()
    ml_recharge.solve()


# Tests for confidence intervals and prediction intervals
def test_pred_interval(ml_recharge: ps.Model) -> None:
    ml_recharge.solve(solver=ps.LeastSquares())
    pi = ml_recharge.solver.prediction_interval(n=10)
    assert isinstance(pi, pd.DataFrame)
    assert pi.shape[1] == 2
    assert list(pi.columns) == [0.025, 0.975]


def test_ci_simulation(ml_recharge: ps.Model) -> None:
    ml_recharge.solve(solver=ps.LeastSquares())
    ci = ml_recharge.solver.ci_simulation(n=10)
    assert isinstance(ci, pd.DataFrame)
    assert ci.shape[1] == 2


def test_ci_block_response(ml_recharge: ps.Model) -> None:
    ml_recharge.solve(solver=ps.LeastSquares())
    ci = ml_recharge.solver.ci_block_response(name="rch", n=10)
    assert isinstance(ci, pd.DataFrame)
    assert ci.shape[1] == 2


def test_ci_step_response(ml_recharge: ps.Model) -> None:
    ml_recharge.solve(solver=ps.LeastSquares())
    ci = ml_recharge.solver.ci_step_response(name="rch", n=10)
    assert isinstance(ci, pd.DataFrame)
    assert ci.shape[1] == 2


def test_ci_contribution(ml_recharge: ps.Model) -> None:
    ml_recharge.solve(solver=ps.LeastSquares())
    ci = ml_recharge.solver.ci_contribution(name="rch", n=10)
    assert isinstance(ci, pd.DataFrame)
    assert ci.shape[1] == 2


# Test the EmceeSolver
def test_emcee(ml_recharge: ps.Model) -> None:
    try:
        ml_recharge.solve(solver=ps.LeastSquares())
        ml_recharge.del_noisemodel()
        ml_recharge.solve(
            solver=ps.EmceeSolve(nwalkers=10),
            initial=False,
            fit_constant=False,
            steps=2,
        )
    except ImportError:
        pytest.skip("emcee not installed, skipping test")


# Unit tests for the BaseSolver class using pytest style
class TestBaseSolver:
    """Test the BaseSolver class."""

    def test_init_empty(self) -> None:
        """Test empty initialization of BaseSolver."""
        solver = BaseSolver()
        assert solver._name == "BaseSolver"
        assert solver.ml is None
        assert solver.pcov is None
        assert solver.nfev is None

    def test_init_with_params(self) -> None:
        """Test initialization with parameters."""
        pcov_data = [[0.1, 0.02], [0.02, 0.2]]
        pcov = DataFrame(pcov_data, index=["p1", "p2"], columns=["p1", "p2"])
        solver = BaseSolver(pcov=pcov, nfev=10)

        assert solver.pcov is pcov
        assert solver.nfev == 10

        # Check correlation matrix calculation
        expected_corr = np.array(
            [
                [1.0, 0.02 / (np.sqrt(0.1) * np.sqrt(0.2))],
                [0.02 / (np.sqrt(0.1) * np.sqrt(0.2)), 1.0],
            ]
        )
        np.testing.assert_almost_equal(solver.pcor.values, expected_corr)

    def test_set_model(self, mock_model: MagicMock) -> None:
        """Test setting a model in the solver."""
        solver = BaseSolver()
        solver.set_model(mock_model)
        assert solver.ml is mock_model

    def test_set_model_raises_warning(self, mock_model: MagicMock) -> None:
        """Test that error is raised when setting a second model."""
        solver = BaseSolver()
        solver.set_model(mock_model)
        with pytest.raises(UserWarning):
            solver.set_model(MagicMock())

    @pytest.mark.parametrize(
        "noise,expected_calls",
        [
            (False, ["residuals"]),
            (True, ["noise", "noise_weights"]),
        ],
    )
    def test_misfit_basic(
        self, mock_model: MagicMock, noise: bool, expected_calls: list[str]
    ) -> None:
        """Test misfit function with different noise settings."""
        solver = BaseSolver()
        solver.ml = mock_model
        mock_model.reset_mock()

        result = solver.misfit(p=np.array([1.0, 2.0]), noise=noise)

        assert isinstance(result, np.ndarray)
        assert len(result) == 100

        # Verify the right methods were called
        for method in expected_calls:
            assert getattr(mock_model, method).called

    def test_misfit_with_weights(self, mock_model: MagicMock) -> None:
        """Test misfit function with weights."""
        solver = BaseSolver()
        solver.ml = mock_model
        weights = Series(
            np.random.uniform(0.5, 1.0, 100), index=mock_model.residuals().index
        )

        weighted_result = solver.misfit(
            p=np.array([1.0, 2.0]), noise=False, weights=weights
        )

        assert isinstance(weighted_result, np.ndarray)
        assert len(weighted_result) == 100

    def test_misfit_returnseparate(self, mock_model: MagicMock) -> None:
        """Test misfit function with returnseparate=True."""
        solver = BaseSolver()
        solver.ml = mock_model

        result = solver.misfit(p=np.array([1.0, 2.0]), noise=False, returnseparate=True)

        assert isinstance(result, tuple)
        assert len(result) == 3
        assert all(isinstance(arr, np.ndarray) for arr in result)

    def test_to_dict(self) -> None:
        """Test to_dict method."""
        solver = BaseSolver(nfev=100)
        data_dict = solver.to_dict()

        assert data_dict["class"] == "BaseSolver"
        assert data_dict["nfev"] == 100
        assert data_dict["pcov"] is None


class TestLeastSquares:
    """Test the LeastSquares solver class."""

    def test_init(self) -> None:
        """Test initialization of LeastSquares solver."""
        solver = LeastSquares()
        assert solver._name == "LeastSquares"

    @patch("pastas.solver.least_squares")
    def test_solve(self, mock_least_squares: MagicMock, mock_model: MagicMock) -> None:
        """Test solve method."""
        # Setup mock
        mock_result = MagicMock(
            success=True,
            x=np.array([1.5, 2.5]),
            cost=10.0,
            nfev=50,
            jac=np.array([[0.1, 0.05], [0.05, 0.2], [0.01, 0.02]]),
        )
        mock_least_squares.return_value = mock_result

        # Create solver and solve
        solver = LeastSquares()
        solver.ml = mock_model
        success, optimal, stderr = solver.solve(noise=True)

        # Assertions
        assert success is True
        assert len(optimal) == 3
        assert np.allclose(optimal[:2], [1.5, 2.5])
        assert isinstance(stderr, np.ndarray)
        assert solver.nfev == 50

    @pytest.mark.parametrize(
        "method,absolute_sigma",
        [("trf", False), ("trf", True), ("lm", False), ("dogbox", False)],
    )
    def test_get_covariances(self, method: str, absolute_sigma: bool) -> None:
        """Test get_covariances with different methods and parameters."""
        jacobian = np.array([[1.0, 0.5], [0.5, 1.0], [0.2, 0.3]])
        cost = 5.0

        pcov = LeastSquares.get_covariances(
            jacobian, cost, method=method, absolute_sigma=absolute_sigma
        )

        assert isinstance(pcov, np.ndarray)
        assert pcov.shape == (2, 2)
        assert not np.isnan(pcov).any()


class TestOptionalSolvers:
    """Tests for solvers that depend on optional dependencies."""

    def test_lmfit_solve_init(self) -> None:
        """Test LmfitSolve initialization."""
        try:
            solver = LmfitSolve()
            assert solver._name == "LmfitSolve"
        except ImportError:
            pytest.skip("lmfit not installed")

    def test_emcee_solve_init(self) -> None:
        """Test EmceeSolve initialization."""
        try:
            solver = EmceeSolve()
            assert solver._name == "EmceeSolve"
            assert solver.nwalkers == 20
            assert solver.progress_bar is True
        except ImportError:
            pytest.skip("emcee not installed")

    def test_emcee_to_dict_raises(self) -> None:
        """Test that EmceeSolve.to_dict raises NotImplementedError."""
        try:
            solver = EmceeSolve()
            with pytest.raises(NotImplementedError):
                solver.to_dict()
        except ImportError:
            pytest.skip("emcee not installed")



from bokeh.layouts import column, row
from bokeh.models import (
    ColumnDataSource,
    DataTable,
    RangeTool,
    ScientificFormatter,
    StringFormatter,
    TableColumn,
)
from bokeh.plotting import figure, show

from pastas.extensions import register_model_accessor


@register_model_accessor("bokeh")
class Bokeh:
    """Extension class for interactive bokeh figures for pastas Models.

    Examples
    --------
    >>> ps.extensions.register_bokeh()
    INFO: Registered bokeh plotting methods in Model class, e.g. `ml.bokeh.plot()`.
    >>> fig = ml.bokeh.results()
    >>> fig.write_html("results_figure.html")

    Methods
    -------
    plot
        plot oseries and model simulation, interactive version of `ml.plot()`
    results
        plot oseries, model simulation, contribution, step responses and parameters
        table,interactive version of `ml.plots.results()`

    Notes
    -----
    The `bokeh` extension is registered in the `Model` class by calling the
    `register_bokeh()` function. To work in Juptyer notebooks, the
    `bokeh.io.output_notebook()` function should be called before plotting. The `bokeh`
    extension is not registered by default, and should be called explicitly. Check the
    bokeh documentation for more information on how to interact with the plots.

    """

    def __init__(self, model):
        self._model = model

    def plot(self, tmin=None, tmax=None, height=300, width=600, show_plot=True):
        """Plot the observations and model simulation.

        Parameters
        ----------
        tmin : pd.Timestamp, optional
            start time for model simulation, by default None
        tmax : pd.Timestamp, optional
            end time for model simulation, by default None
        height : int, optional
             height of the plot, by default 500
        width : int, optional
           width of the plot, by default 800
        show_plot : bool, optional
            Show the plot (i.e., in Jupyter Notebooks), by default True

        Returns
        -------
        p : bokeh.plotting.figure
            Bokeh figure with the observations and model simulation.

        Examples
        --------
        >>> ps.extensions.register_bokeh()
        INFO: Registered bokeh plotting methods in Model class, e.g. `ml.bokeh.plot()`.
        >>>
        >>> fig = ml.bokeh.plot()

        """

        data = self._model.get_output_series(tmin=tmin, tmax=tmax, split=False)
        source = ColumnDataSource(data)
        rsq = self._model.stats.rsq(tmin=tmin, tmax=tmax)

        TOOLS = "zoom_in,zoom_out,reset,pan,xwheel_zoom,box_zoom,undo"

        p = figure(
            title="Pastas Model",
            y_axis_label="Head",
            x_axis_location=None,
            tools=TOOLS,
            width=width,
            height=height,
            x_axis_type="datetime",
            toolbar_location="above",
        )

        p.scatter(
            "index",
            "Head_Calibration",
            source=source,
            legend_label="Observations",
            color="black",
            alpha=0.7,
        )
        p.line(
            "index",
            "Simulation",
            source=source,
            legend_label=r"Simulation (R2 = {:.2f})".format(rsq),
            line_width=2,
        )
        p.legend.ncols = 2

        if show_plot:
            show(p)
        return p

    def results(
        self, tmin=None, tmax=None, height=500, width=800, show_plot=True, stderr=False
    ):
        """Overview of the results of the pastas model.

        Parameters
        ----------
        tmin : pd.Timestamp, optional
            start time for model simulation, by default None
        tmax : pd.Timestamp, optional
            end time for model simulation, by default None
        height : int, optional
             height of the plot, by default 500
        width : int, optional
           width of the plot, by default 800
        show_plot : bool, optional
            Show the plot (i.e., in Jupyter Notebooks), by default True
        stderr : bool, optional
            Show the standard error of the parameters, by default False

        Returns
        -------
        grid : bokeh.layouts.column
            Bokeh layout with the results of the pastas model.

        Examples
        --------
        >>> ps.extensions.register_bokeh()
        INFO: Registered bokeh plotting methods in Model class, e.g. `ml.bokeh.plot()`.
        >>> fig = ml.bokeh.results()

        """
        data = self._model.get_output_series(tmin=tmin, tmax=tmax, split=False)
        ranges = data.max() - data.min()

        ranges = ranges.drop(ranges.iloc[:2].idxmin())

        if self._model.settings["noise"]:
            ranges = ranges.drop("Noise")

        heights = (ranges / ranges.sum() * (height - 50)).values.astype(int)
        source = ColumnDataSource(data)
        rsq = self._model.stats.rsq(tmin=tmin, tmax=tmax)

        TOOLS = "zoom_in,zoom_out,reset,pan,xwheel_zoom,box_zoom,undo"

        p = figure(
            title="Pastas Model",
            y_axis_label="Head",
            x_axis_location=None,
            tools=TOOLS,
            width=int(0.75 * width),
            height=heights[0],
            x_axis_type="datetime",
            toolbar_location="above",
        )

        p.scatter(
            "index",
            "Head_Calibration",
            source=source,
            legend_label="Observations",
            color="black",
            alpha=0.7,
        )
        p.line(
            "index",
            "Simulation",
            source=source,
            legend_label=r"Simulation (R2 = {:.2f})".format(rsq),
        )
        p.legend.ncols = 2

        # Residuals
        res_plot = figure(
            y_axis_label="Residuals",
            toolbar_location=None,
            tools=TOOLS,
            x_range=p.x_range,
            width=int(0.75 * width),
            height=heights[2],
            x_axis_type="datetime",
            x_axis_location=None,
        )

        res_plot.scatter(
            "index",
            "Residuals",
            source=source,
            color="black",
            alpha=0.7,
            legend_label="Residuals",
        )
        res_plot.line(
            "index",
            "Residuals",
            source=source,
            color="black",
            alpha=0.7,
            legend_label="Residuals",
        )

        if self._model.settings["noise"]:
            res_plot.line("index", "Noise", source=source, legend_label="Noise")
            res_plot.scatter("index", "Noise", source=source, legend_label="Noise")

        res_plot.legend.ncols = 2

        # Parameter Table
        df = ColumnDataSource(self._model.parameters.loc[:, ["optimal", "stderr"]])
        columns = [
            TableColumn(
                field="index",
                title="Name",
                formatter=StringFormatter(font_style="bold"),
            ),
            TableColumn(
                field="optimal",
                title="Optimal",
                formatter=ScientificFormatter(precision=2),
            ),
        ]
        if stderr:
            columns.append(
                TableColumn(
                    field="stderr",
                    title="Stderr",
                    formatter=ScientificFormatter(precision=2),
                )
            )

        table = DataTable(
            source=df,
            columns=columns,
            editable=False,
            index_position=None,
            width=int(0.25 * width),
            height=heights[0] + heights[2] - 10,
        )

        left_column = [p, res_plot]
        right_column = [table]

        # Contributions
        rfunc_plot = None

        for i, smname in enumerate(self._model.stressmodels.keys(), start=2):
            if i == int(len(self._model.stressmodels) + 1):
                x_axis_location = "below"
            else:
                x_axis_location = None

            contrib_plot = figure(
                y_axis_label="Rise",
                toolbar_location=None,
                tools=TOOLS,
                x_axis_location=x_axis_location,
                width=int(0.75 * width),
                height=heights[i],
                x_axis_type="datetime",
                x_range=p.x_range,
            )

            # xrange = False if rfunc_plot is None else rfunc_plot.x_range

            rfunc_plot = figure(
                x_axis_label=None,
                toolbar_location=None,
                x_axis_location=x_axis_location,
                width=int(0.25 * width),
                height=heights[i],
            )

            contrib_plot.line("index", smname, source=source)
            response = self._model.get_step_response(smname)
            rfunc_plot.line(response.index, response.values)

            left_column.append(contrib_plot)
            right_column.append(rfunc_plot)

        select = figure(
            height=50,
            width=int(0.75 * width),
            y_range=p.y_range,
            x_axis_type="datetime",
            y_axis_type=None,
            tools="",
            toolbar_location=None,
            background_fill_color="#ffffff",
        )

        range_tool = RangeTool(x_range=p.x_range)
        select.line("index", "Simulation", source=source)
        select.add_tools(range_tool)
        left_column.append(select)

        layout = row(column(left_column), column(right_column))
        grid = column(layout, width=width, height=height)

        if show_plot:
            show(grid)

        return grid



"""The following methods may be used to describe the fit between the model simulation
and the observations.

.. currentmodule:: pastas.modelstats.Statistics

.. autosummary::
   :nosignatures:
   :toctree: ./generated

   summary

Examples
========
These methods may be used as follows.


    >>> ml.stats.summary(stats=["rmse", "mae", "nse"])
                  Value
    Statistic
    rmse       0.114364
    mae        0.089956
    nse        0.929136
"""

from numpy import nan
from pandas import DataFrame

from pastas.typing import Model, TimestampType

from .decorators import model_tmin_tmax
from .stats import diagnostics, metrics


class Statistics:
    # Save all statistics that can be calculated.
    ops = [
        "rmse",
        "rmsn",
        "sse",
        "mae",
        "nse",
        "evp",
        "rsq",
        "kge",
        "bic",
        "aic",
        "aicc",
    ]

    def __init__(self, ml: Model):
        """This class provides statistics to pastas Model class.

        Parameters
        ----------
        ml: Pastas.model.Model
            ml is a time series Model that is calibrated.

        Notes
        -----
        To obtain a list of all statistics that are included type:

        >>> print(ml.stats.ops)
        """
        # Save a reference to the model.
        self.ml = ml

    def __repr__(self):
        msg = """This module contains all the statistical functions included in Pastas.

        To obtain a list of all statistics that are included type:

    >>> print(ml.stats.ops)"""
        return msg

    @model_tmin_tmax
    def rmse(
        self,
        tmin: TimestampType | None = None,
        tmax: TimestampType | None = None,
        weighted: bool = False,
        **kwargs,
    ) -> float:
        """Root mean squared error of the residuals.

        Parameters
        ----------
        tmin: str or pandas.Timestamp, optional
        tmax: str or pandas.Timestamp, optional
        weighted: bool, optional
            If weighted is True, the variances are computed using the time step
            between observations as weights. Default is False.

        See Also
        --------
        pastas.stats.rmse
        """
        res = self.ml.residuals(tmin=tmin, tmax=tmax)
        return metrics.rmse(res=res, weighted=weighted, **kwargs)

    @model_tmin_tmax
    def rmsn(
        self,
        tmin: TimestampType | None = None,
        tmax: TimestampType | None = None,
        weighted: bool = False,
        **kwargs,
    ) -> float:
        """Root mean squared error of the noise.

        Parameters
        ----------
        tmin: str or pandas.Timestamp, optional
        tmax: str or pandas.Timestamp, optional
        weighted: bool, optional
            If weighted is True, the variances are computed using the time step
            between observations as weights. Default is False.

        Returns
        -------
        float or nan
            Return a float if noisemodel is present, nan if not.

        See Also
        --------
        pastas.stats.rmse
        """
        if not self.ml.settings["noise"]:
            return nan
        else:
            res = self.ml.noise(tmin=tmin, tmax=tmax)
            return metrics.rmse(res=res, weighted=weighted, **kwargs)

    @model_tmin_tmax
    def sse(
        self, tmin: TimestampType | None = None, tmax: TimestampType | None = None
    ) -> float:
        """Sum of the squares of the error (SSE)

        Parameters
        ----------
        tmin: str or pandas.Timestamp, optional
        tmax: str or pandas.Timestamp, optional

        See Also
        --------
        pastas.stats.sse
        """
        res = self.ml.residuals(tmin=tmin, tmax=tmax)
        return metrics.sse(res=res)

    @model_tmin_tmax
    def mae(
        self,
        tmin: TimestampType | None = None,
        tmax: TimestampType | None = None,
        weighted: bool = False,
        **kwargs,
    ) -> float:
        """Mean Absolute Error (MAE) of the residuals.

        Parameters
        ----------
        tmin: str or pandas.Timestamp, optional
        tmax: str or pandas.Timestamp, optional
        weighted: bool, optional
            If weighted is True, the variances are computed using the time step
            between observations as weights. Default is False.

        See Also
        --------
        pastas.stats.mae
        """
        res = self.ml.residuals(tmin=tmin, tmax=tmax)
        return metrics.mae(res=res, weighted=weighted, **kwargs)

    @model_tmin_tmax
    def nse(
        self,
        tmin: TimestampType | None = None,
        tmax: TimestampType | None = None,
        weighted: bool = False,
        **kwargs,
    ) -> float:
        """Nash-Sutcliffe Efficiency for model fit .

        Parameters
        ----------
        tmin: str or pandas.Timestamp, optional
        tmax: str or pandas.Timestamp, optional
        weighted: bool, optional
            If weighted is True, the variances are computed using the time step
            between observations as weights. Default is False.

        See Also
        --------
        pastas.stats.nse
        """
        res = self.ml.residuals(tmin=tmin, tmax=tmax)
        obs = self.ml.observations(tmin=tmin, tmax=tmax)
        return metrics.nse(obs=obs, res=res, weighted=weighted, **kwargs)

    @model_tmin_tmax
    def nnse(
        self,
        tmin: TimestampType | None = None,
        tmax: TimestampType | None = None,
        weighted: bool = False,
        **kwargs,
    ) -> float:
        """Normalized Nash-Sutcliffe Efficiency for model fit .

        Parameters
        ----------
        tmin: str or pandas.Timestamp, optional
        tmax: str or pandas.Timestamp, optional
        weighted: bool, optional
            If weighted is True, the variances are computed using the time step
            between observations as weights. Default is False.

        See Also
        --------
        pastas.stats.nnse
        """
        res = self.ml.residuals(tmin=tmin, tmax=tmax)
        obs = self.ml.observations(tmin=tmin, tmax=tmax)

        return metrics.nnse(obs=obs, res=res, weighted=weighted, **kwargs)

    @model_tmin_tmax
    def pearsonr(
        self,
        tmin: TimestampType | None = None,
        tmax: TimestampType | None = None,
        weighted: bool = False,
        **kwargs,
    ) -> float:
        """Compute the (weighted) Pearson correlation (r).

        Parameters
        ----------
        tmin: str or pandas.Timestamp, optional
        tmax: str or pandas.Timestamp, optional
        weighted: bool, optional
            If weighted is True, the variances are computed using the time step
            between observations as weights. Default is False.

        See Also
        --------
        pastas.stats.pearsonr
        """
        obs = self.ml.observations(tmin=tmin, tmax=tmax)
        sim = self.ml.simulate(tmin=tmin, tmax=tmax)
        return metrics.pearsonr(obs=obs, sim=sim, weighted=weighted, **kwargs)

    @model_tmin_tmax
    def evp(
        self,
        tmin: TimestampType | None = None,
        tmax: TimestampType | None = None,
        weighted: bool = False,
        **kwargs,
    ) -> float:
        """Explained variance percentage.

        Parameters
        ----------
        tmin: str or pandas.Timestamp, optional
        tmax: str or pandas.Timestamp, optional
        weighted: bool, optional
            If weighted is True, the variances are computed using the time step
            between observations as weights. Default is False.

        See Also
        --------
        pastas.stats.evp
        """
        res = self.ml.residuals(tmin=tmin, tmax=tmax)
        obs = self.ml.observations(tmin=tmin, tmax=tmax)
        return metrics.evp(obs=obs, res=res, weighted=weighted, **kwargs)

    @model_tmin_tmax
    def rsq(
        self,
        tmin: TimestampType | None = None,
        tmax: TimestampType | None = None,
        weighted: bool = False,
        **kwargs,
    ) -> float:
        """R-squared.

        Parameters
        ----------
        tmin: str or pandas.Timestamp, optional
        tmax: str or pandas.Timestamp, optional
        weighted: bool, optional
            If weighted is True, the variances are computed using the time step
            between observations as weights. Default is False.

        See Also
        --------
        pastas.stats.rsq
        """
        obs = self.ml.observations(tmin=tmin, tmax=tmax)
        res = self.ml.residuals(tmin=tmin, tmax=tmax)
        return metrics.rsq(obs=obs, res=res, weighted=weighted, **kwargs)

    @model_tmin_tmax
    def kge(
        self,
        tmin: TimestampType | None = None,
        tmax: TimestampType | None = None,
        weighted: bool = False,
        modified: bool = False,
        **kwargs,
    ) -> float:
        """Kling-Gupta Efficiency.

        Parameters
        ----------
        tmin: str or pandas.Timestamp, optional
        tmax: str or pandas.Timestamp, optional
        weighted: bool, optional
            If weighted is True, the variances are computed using the time step
            between observations as weights. Default is False.
        modified: bool, optional
            Use the modified KGE as proposed by :cite:t:`kling_runoff_2012`.

        See Also
        --------
        pastas.stats.kge
        """
        sim = self.ml.simulate(tmin=tmin, tmax=tmax)
        obs = self.ml.observations(tmin=tmin, tmax=tmax)
        return metrics.kge(
            obs=obs, sim=sim, weighted=weighted, modified=modified, **kwargs
        )

    @model_tmin_tmax
    def kge_2012(
        self,
        tmin: TimestampType | None = None,
        tmax: TimestampType | None = None,
        weighted: bool = False,
        **kwargs,
    ) -> float:
        """Kling-Gupta Efficiency.

        Parameters
        ----------
        tmin: str or pandas.Timestamp, optional
        tmax: str or pandas.Timestamp, optional
        weighted: bool, optional
            If weighted is True, the variances are computed using the time step
            between observations as weights. Default is False.

        See Also
        --------
        pastas.stats.kge_2012
        """
        sim = self.ml.simulate(tmin=tmin, tmax=tmax)
        obs = self.ml.observations(tmin=tmin, tmax=tmax)
        return metrics.kge_2012(obs=obs, sim=sim, weighted=weighted, **kwargs)

    @model_tmin_tmax
    def bic(
        self, tmin: TimestampType | None = None, tmax: TimestampType | None = None
    ) -> float:
        """Bayesian Information Criterium (BIC).

        Parameters
        Parameters
        ----------
        tmin: str or pandas.Timestamp, optional
        tmax: str or pandas.Timestamp, optional

        See Also
        --------
        pastas.stats.bic
        """
        nparam = self.ml.parameters["vary"].sum()
        if self.ml.settings["noise"]:
            res = self.ml.noise(tmin=tmin, tmax=tmax) * self.ml.noise_weights(
                tmin=tmin, tmax=tmax
            )
        else:
            res = self.ml.residuals(tmin=tmin, tmax=tmax)
        return metrics.bic(res=res, nparam=nparam)

    @model_tmin_tmax
    def aic(
        self, tmin: TimestampType | None = None, tmax: TimestampType | None = None
    ) -> float:
        """Akaike Information Criterium (AIC).

        Parameters
        ----------
        tmin: str or pandas.Timestamp, optional
        tmax: str or pandas.Timestamp, optional

        See Also
        --------
        pastas.stats.aic
        """
        nparam = self.ml.parameters["vary"].sum()
        if self.ml.settings["noise"]:
            res = self.ml.noise(tmin=tmin, tmax=tmax) * self.ml.noise_weights(
                tmin=tmin, tmax=tmax
            )
        else:
            res = self.ml.residuals(tmin=tmin, tmax=tmax)
        return metrics.aic(res=res, nparam=nparam)

    @model_tmin_tmax
    def aicc(
        self, tmin: TimestampType | None = None, tmax: TimestampType | None = None
    ) -> float:
        """Akaike Information Criterium with second order bias correction (AICc).

        Parameters
        ----------
        tmin: str or pandas.Timestamp, optional
        tmax: str or pandas.Timestamp, optional

        See Also
        --------
        pastas.stats.aicc
        """
        nparam = self.ml.parameters["vary"].sum()
        if self.ml.settings["noise"]:
            res = self.ml.noise(tmin=tmin, tmax=tmax) * self.ml.noise_weights(
                tmin=tmin, tmax=tmax
            )
        else:
            res = self.ml.residuals(tmin=tmin, tmax=tmax)
        return metrics.aicc(res=res, nparam=nparam)

    @model_tmin_tmax
    def summary(
        self,
        tmin: TimestampType | None = None,
        tmax: TimestampType | None = None,
        stats: list[str] | None = None,
    ) -> DataFrame:
        """Returns a Pandas DataFrame with goodness-of-fit metrics.

        Parameters
        ----------
        tmin: str or pandas.Timestamp, optional
        tmax: str or pandas.Timestamp, optional
        stats: list, optional
            list of statistics that need to be calculated. If nothing is provided,
            all statistics are returned.

        Returns
        -------
        stats : Pandas.DataFrame
            single-column DataFrame with calculated statistics.

        Examples
        --------

        >>> ml.stats.summary()

        or

        >>> ml.stats.summary(stats=["mae", "rmse"])
        """

        if stats is None:
            stats_to_compute = self.ops
        else:
            stats_to_compute = stats

        # Remove rmsn if no noise model
        if "rmsn" in stats_to_compute and not (
            self.ml.noisemodel and self.ml.settings["noise"]
        ):
            stats_to_compute.remove("rmsn")

        stats = DataFrame(columns=["Value"])

        for k in stats_to_compute:
            stats.loc[k] = getattr(self, k)(tmin=tmin, tmax=tmax)

        stats.index.name = "Statistic"
        return stats

    @model_tmin_tmax
    def diagnostics(
        self,
        tmin: TimestampType | None = None,
        tmax: TimestampType | None = None,
        alpha: float = 0.05,
        stats: tuple = (),
        float_fmt: str = "{0:.2f}",
    ) -> DataFrame:
        """Methods to compute various diagnostics checks for the noise time series. If
        no NoiseModel is used, the diagnostics are computed on the model residuals.

        Parameters
        ----------
        tmin: str or pandas.Timestamp, optional
            Start date of the residual / noise time series to compute the diagnostics
            checks for.
        tmax: str or pandas.Timestamp, optional
            End date of the residuals / noise time series to compute the diagnostics
            checks for.
        alpha: float, optional
            significance level to use for the hypothesis testing.
        stats: tuple, optional
            Tuple with the diagnostic checks to perform. Not implemented yet.
        float_fmt: str
            String to use for formatting the floats in the returned DataFrame.

        Returns
        -------
        df: Pandas.DataFrame
            DataFrame with the information for the diagnostics checks. The final column
            in this DataFrame report if the Null-Hypothesis is rejected. If H0 is not
            rejected (=False) the data is in agreement with one of the properties of
            white noise (e.g., normally distributed).

        See Also
        --------
        pastas.stats.diagnostics

        """
        if self.ml.noisemodel and self.ml.settings["noise"]:
            series = self.ml.noise(tmin=tmin, tmax=tmax)
            nparam = self.ml.noisemodel.nparam
        else:
            series = self.ml.residuals(tmin=tmin, tmax=tmax)
            nparam = 0

        return diagnostics(
            series=series, alpha=alpha, nparam=nparam, stats=stats, float_fmt=float_fmt
        )



"""This module contains a timer for model optimization.

The timer prints the time elapsed and number of iterations. Optionally, a maximum solve
time can be specified, to abort long optimizations. This class is not automatically
imported in pastas and requires the tqdm module (pip install tqdm).

Usage::

    from pastas.timer import SolveTimer

    with SolveTimer(max_time=60) as t:  # max time in seconds
        ml.solve(callback=t.timer)

This will print the following to the console::

    Optimization progress: 73it [00:01, 67.68it/s]

"""

try:
    from tqdm.auto import tqdm
except ImportError:
    msg = "SolveTimer requires 'tqdm' to be installed."
    raise ImportError(msg) from None


class ExceededMaxSolveTime(Exception):
    """Custom Exception when model optimization exceeds threshold."""


class SolveTimer(tqdm):
    """Progress indicator for model optimization.

    Examples
    --------
    Print timer and number of iterations in console while running `ml.solve()`::

        with SolveTimer() as t:
            ml.solve(callback=t.timer)

    This prints the following to the console, for example::

        Optimization progress: 73it [00:01, 67.68it/s]

    Set maximum allowable time (in seconds) for solve, otherwise raise
    ExceededMaxSolveTime exception::

        with SolveTimer(max_time=60) as t:
            ml.solve(callback=t.timer)

    Notes
    -----
    If the logger is also printing messages to the console the timer will not be
    updated quite as nicely.
    """

    def __init__(self, max_time: float | None = None, *args, **kwargs) -> None:
        """Initialize SolveTimer.

        Parameters
        ----------
        max_time : float, optional
            maximum allowed time spent in solve(), by default None, which does
            not impose a limit. If time is exceeded, raises
            ExceededMaxSolveTime Exception.
        """
        if "total" not in kwargs:
            kwargs["total"] = None
        if "desc" not in kwargs:
            kwargs["desc"] = "Optimization progress"
        self.max_time = max_time
        super().__init__(*args, **kwargs)

    def timer(self, _, n: int = 1):
        """Callback method for ps.Model.solve()."""
        displayed = super().update(n)
        if self.max_time is not None:
            if self.format_dict["elapsed"] > self.max_time:
                raise ExceededMaxSolveTime(
                    f"Model solve time exceeded {self.max_time} seconds!"
                )
        return displayed



"""This module contains plotting methods for Pastas."""

import logging

import matplotlib.patheffects as path_effects
import matplotlib.pyplot as plt
import numpy as np
from pandas import DataFrame, Series, Timestamp, concat
from scipy.stats import gaussian_kde, norm, pearsonr, probplot

from pastas.plotting.modelcompare import CompareModels
from pastas.plotting.plotutil import plot_series_with_gaps, share_xaxes, share_yaxes
from pastas.stats.core import acf as get_acf
from pastas.stats.metrics import evp, rmse
from pastas.typing import ArrayLike, Axes, Figure, Model, TimestampType

logger = logging.getLogger(__name__)

__all__ = ["compare", "series", "acf", "diagnostics", "cum_frequency", "TrackSolve"]


def compare(
    models: list[Model],
    names: list[str] | None = None,
    adjust_height: bool = True,
    tmin: TimestampType | None = None,
    tmax: TimestampType | None = None,
    **kwargs,
) -> dict:
    """Plot multiple Pastas models in one figure to visually compare models.

    Notes
    -----
    The models must have the same stressmodel names, otherwise the contributions will
    not be plotted, and parameters table will not display nicely.

    Parameters
    ----------
    models: list
        List of pastas Models, works for N models, but certain things might not
        display nicely if the list gets too long.
    names : list of str
        override model names by passing a list of names
    adjust_height: bool, optional
        Adjust the height of the graphs, so that the vertical scale of all the
        subplots on the left is equal. Default is False, in which case the axes are
        not rescaled to include all data, so certain data might not be visible. Set
        False to ensure you can see all data.
    tmin: TimestampType, optional
        Timestamp with a start date for the simulation period (E.g. '1980'). If none
        is provided, the tmin from the oseries is used.
    tmax: TimestampType, optional
        Timestamp with an end date for the simulation period (E.g. '2010'). If none
        is provided, the tmax from the oseries is used.
    **kwargs
        The kwargs are passed to the CompareModels.plot() function.

    Returns
    -------
    matplotlib.axes
    """
    mc = CompareModels(models, names=names, tmin=tmin, tmax=tmax)
    mc.plot(adjust_height=adjust_height, **kwargs)
    return mc.axes


def series(
    head: Series | None = None,
    stresses: list[Series] | None = None,
    hist: bool = True,
    kde: bool = False,
    table: bool = False,
    titles: bool = True,
    tmin: TimestampType | None = None,
    tmax: TimestampType | None = None,
    colors_stresses: list[str] | None = None,
    labels: list[str] | None = None,
    figsize: tuple = (10, 5),
    **kwargs,
) -> Axes:
    """Plot all the input time Series in a single plot.

    Parameters
    ----------
    head: pd.Series
        Pandas time series with DatetimeIndex.
    stresses: list of pd.Series
        List with Pandas time series with DatetimeIndex.
    hist: bool
        Histogram for the series. The number of bins is determined with Sturges rule.
    kde: bool
        Kernel density estimate for the series. The kde is obtained from
        scipy.gaussian_kde using scott to calculate the estimator bandwidth. Returns
        the number of observations, mean, skew and kurtosis.
    table: bool
        Show table with some basic statistics such as the number of
        observations, mean, skew and kurtosis.
    titles: bool
        Set the titles or not. Taken from the name attribute of the series.
    tmin: str or pd.Timestamp
    tmax: str or pd.Timestamp
    colors_stresses: list of str
        List with the matplotlib colorcodes to use for plotting each stress timeseries.
        If list is shorter than number of stresses, the remaining stresses are plotted
        in black. If None (default), default matplotlib colors will be used.
    labels: list of str
        List with the labels for each subplot.
    figsize: tuple
        Set the size of the figure.
    kwargs:
        keyword arguments passed to plotting functions of stress timeseries

    Returns
    -------
    matplotlib.Axes
    """
    nrows = 0
    if head is not None:
        nrows += 1
        tmin = head.index[0] if tmin is None else tmin
        tmax = head.index[-1] if tmax is None else tmax
    if stresses is not None:
        nrows += len(stresses)
    if colors_stresses is None:
        colors_stresses = plt.rcParams["axes.prop_cycle"].by_key()["color"]
    gridspec_kw = {}
    cols = 1
    if table and not hist and not kde:
        logging.info(
            "Plotting the table is not possible without hist=True or kde=True. Adding the histogram."
        )
        hist = True
    if hist or kde:
        gridspec_kw["width_ratios"] = [3, 1]
        cols += 1
        if table:
            cols += 1
            gridspec_kw["width_ratios"].append(1)
    _, axes = plt.subplots(
        nrows,
        cols,
        figsize=figsize,
        sharey="row",
        gridspec_kw=gridspec_kw,
    )
    if nrows == 1 and cols == 1:
        axes = np.array([[axes]])
    elif nrows == 1:
        axes = axes[np.newaxis]
    elif cols == 1:
        axes = axes[:, np.newaxis]
    if hist:
        axes[-1, 1].set_xlabel("Frequency [%]")
    if kde:
        axes[-1, 1].set_xlabel("Density [-]")
    if head is not None:
        head.plot(
            ax=axes[0, 0], marker=".", linestyle=" ", color="k", xlabel="", **kwargs
        )
        if titles:
            axes[0, 0].set_title(head.name)
        if labels is not None:
            axes[0, 0].set_ylabel(labels[0])
        if hist:
            weights = None if kde else np.ones(len(head)) / len(head) * 100
            head.hist(
                ax=axes[0, 1],
                orientation="horizontal",
                color="k",
                weights=weights,
                bins=int(np.ceil(1 + np.log2(len(head)))),
                grid=False,
                density=kde,
            )
        if kde:
            gkde = gaussian_kde(head, bw_method="scott")
            sample_range = np.max(head) - np.min(head)
            ind = np.linspace(
                np.min(head) - 0.1 * sample_range,
                np.max(head) + 0.1 * sample_range,
                1000,
            )
            color = "darkgrey" if hist else "k"
            axes[0, 1].plot(gkde.evaluate(ind), ind, color=color)
        if table:
            # stats table
            head_stats = [
                ["Count", f"{head.count():0.0f}"],
                ["Mean", f"{head.mean():0.2f}"],
                ["Max", f"{head.max():0.2f}"],
                ["Min", f"{head.min():0.2f}"],
                ["Skew", f"{head.skew():0.2f}"],
                ["Kurtosis", f"{head.kurtosis():0.2f}"],
            ]
            axes[0, 2].table(
                bbox=(0.0, 0.0, 1, 1), colWidths=(1.5, 1), cellText=head_stats
            )
            axes[0, 2].axis("off")

    if stresses is not None:
        for i, stress in enumerate(stresses, start=nrows - len(stresses)):
            stress = stress[tmin:tmax].dropna()
            if i <= len(colors_stresses):
                color_stress = colors_stresses[i - 1]
            else:
                color_stress = "k"
            stress.plot(ax=axes[i, 0], color=color_stress, xlabel="", **kwargs)
            if titles:
                axes[i, 0].set_title(stress.name)
            if labels is not None:
                axes[i, 0].set_ylabel(labels[i])
            if hist:
                weights = None if kde else np.ones(len(stress)) / len(stress) * 100
                stress.hist(
                    ax=axes[i, 1],
                    orientation="horizontal",
                    color=color_stress,
                    weights=weights,
                    bins=int(np.ceil(1 + np.log2(len(stress)))),
                    grid=False,
                    density=kde,
                )
            if kde:
                gkde = gaussian_kde(stress, bw_method="scott")
                sample_range = np.max(stress) - np.min(stress)
                ind = np.linspace(
                    np.min(stress) - 0.1 * sample_range,
                    np.max(stress) + 0.1 * sample_range,
                    1000,
                )
                color = "darkgrey" if hist else color_stress
                axes[i, 1].plot(gkde.evaluate(ind), ind, color=color)
            if table:
                # stats table
                stress_stats = [
                    ["Count", f"{stress.count():0.0f}"],
                    ["Mean", f"{stress.mean():0.2f}"],
                    ["Skew", f"{stress.skew():0.2f}"],
                    ["Kurtosis", f"{stress.kurtosis():0.2f}"],
                ]
                axes[i, 2].table(
                    bbox=(0, 0, 1, 1), colWidths=(1.5, 1), cellText=stress_stats
                )
                axes[i, 2].axis("off")

    share_xaxes(axes[:, 0])

    return axes


def acf(
    series: Series,
    alpha: float = 0.05,
    lags: int = 365,
    acf_options: dict | None = None,
    smooth_conf: bool = True,
    color: str = "k",
    ax: Axes | None = None,
    figsize: tuple = (5, 2),
) -> Axes:
    """Plot of the autocorrelation function of a time series.

    Parameters
    ----------
    series: pandas.Series
        Residual series to plot the autocorrelation function for.
    alpha: float, optional
        Significance level to calculate the (1-alpha)-confidence intervals. For 95%
        confidence intervals, alpha should be 0.05.
    lags: int, optional
        Maximum number of lags (in days) to compute the autocorrelation for.
    acf_options: dict, optional
        Dictionary with keyword arguments passed on to pastas.stats.acf.
    smooth_conf: bool, optional
        For irregular time series the confidence interval may be.
    color: str, optional
        Color of the vertical autocorrelation lines.
    ax: matplotlib.axes.Axes, optional
        Matplotlib Axes instance to plot the ACF on. A new Figure and Axes is created
        when no value for ax is provided.
    figsize: tuple, optional
        2-D Tuple to determine the size of the figure created. Ignored if ax is also
        provided.

    Returns
    -------
    ax: matplotlib.axes.Axes

    Examples
    --------
    >>> res = pd.Series(index=pd.date_range(start=0, periods=1000, freq="D"),
    >>>                 data=np.random.rand(1000))
    >>> ps.plots.acf(res)

    Raises
    ------
    Warning if the ACF is empty. The plot will still be created to ensure that scripts
    will still run when dealing with many models.

    """
    if ax is None:
        _, ax = plt.subplots(1, 1, figsize=figsize)

    # Plot the autocorrelation
    if acf_options is None:
        acf_options = {}
    r = get_acf(series, full_output=True, alpha=alpha, lags=lags, **acf_options)

    if r.empty:
        # Raise a warning
        logger.warning(
            "The computed autocorrelation function has no values. Changing the input "
            "arguments ('acf_options') for calculating ACF may help. No data will be "
            "plotted."
        )
    else:
        if smooth_conf:
            conf = r.conf.rolling(10, min_periods=1).mean().values
        else:
            conf = r.conf.values

        ax.fill_between(r.index.days, conf, -conf, alpha=0.3)
        ax.vlines(r.index.days, [0], r.loc[:, "acf"].values, color=color)
        ax.set_xlim(0, r.index.days.max())

    ax.set_xlabel("Lag [Days]")
    ax.set_ylabel("Autocorrelation [-]")
    ax.set_title("Autocorrelation plot")

    ax.grid(True)
    return ax


def diagnostics(
    series: Series,
    sim: Series | None = None,
    alpha: float = 0.05,
    bins: int = 50,
    acf_options: dict | None = None,
    figsize: tuple = (10, 5),
    fig: Figure | None = None,
    heteroscedasicity: bool = True,
    **kwargs,
) -> Axes:
    """Plot that helps in diagnosing basic model assumptions.

    Parameters
    ----------
    series: pandas.Series
        Pandas Series with the residual time series to diagnose.
    sim: pandas.Series, optional
        Pandas series with the simulated time series. Used to diagnose on
        heteroscedasticity. Ignored if heteroscedasticity is set to False.
    alpha: float, optional
        Significance level to calculate the (1-alpha)-confidence intervals.
    bins: int optional
        Number of bins used for the histogram. 50 is default.
    acf_options: dict, optional
        Dictionary with keyword arguments passed on to pastas.stats.acf.
    figsize: tuple, optional
        Tuple with the height and width of the figure in inches.
    fig: Matplotib.Figure instance, optional
        Optionally provide a Matplotib.Figure instance to plot onto.
    heteroscedasicity: bool, optional
        Create two additional subplots to check for heteroscedasticity. If true,
        a simulated time series has to be provided with the sim argument.
    **kwargs: dict, optional
        Optional keyword arguments, passed on to plt.figure.

    Returns
    -------
    axes: matplotlib.axes.Axes

    Examples
    --------
    >>> res = pd.Series(index=pd.date_range(start=0, periods=1000, freq="D"),
    >>>                 data=np.random.normal(0, 1, 1000))
    >>> ps.stats.plot_diagnostics(res)

    Notes
    -----
    The two right-hand side plots assume that the noise or residuals follow a Normal
    distribution.

    See Also
    --------
    pastas.stats.acf
        Method that computes the autocorrelation.
    scipy.stats.probplot
        Method use to plot the probability plot.
    """
    # Create the figure and axes
    if fig is None:
        fig = plt.figure(figsize=figsize, constrained_layout=True, **kwargs)

    if heteroscedasicity:
        if sim is None:
            msg = (
                "A simulated time series has to be provided to make plots to "
                "diagnose heteroscedasticity. Provide 'sim' argument."
            )
            logger.error(msg=msg)
            raise KeyError(msg)

        gs = fig.add_gridspec(ncols=3, nrows=2, width_ratios=[3, 1, 1])
        ax4 = fig.add_subplot(gs[0, 2])
        ax5 = fig.add_subplot(gs[1, 2])
    else:
        gs = fig.add_gridspec(ncols=2, nrows=2, width_ratios=[3, 1])
    ax = fig.add_subplot(gs[0, 0])
    ax2 = fig.add_subplot(gs[0, 1])
    ax1 = fig.add_subplot(gs[1, 0])
    ax3 = fig.add_subplot(gs[1, 1])

    # Plot the residuals or noise series
    ax.axhline(0, c="k")
    ax = plot_series_with_gaps(series, ax=ax)
    ax.set_ylabel(series.name)
    ax.set_xlim(series.index.min(), series.index.max())
    ax.set_title(f"{series.name} (n={series.size:.0f}, $\\mu$={series.mean():.2f})")
    ax.grid()
    ax.tick_params(axis="x", labelrotation=0)
    for label in ax.get_xticklabels():
        label.set_horizontalalignment("center")

    # Plot the autocorrelation
    acf(series, alpha=alpha, acf_options=acf_options, ax=ax1)
    ax1.set_title(None)

    # Plot the histogram for normality and add a 'best fit' line
    _, bins, _ = ax2.hist(series.values, bins=bins, density=True)
    y = norm.pdf(bins, series.mean(), series.std())
    ax2.plot(bins, y, "k--")
    ax2.set_ylabel("Probability density")
    ax2.set_title("Histogram")

    # Plot the probability plot
    _, (_, _, r) = probplot(series, plot=ax3, dist="norm", rvalue=False)
    c = ax.get_lines()[1].get_color()
    ax3.get_lines()[0].set_color(c)
    ax3.get_lines()[1].set_color("k")

    # Plot R2 here because probplot has suboptimal positioning
    ax3.text(0.5, 0.1, "$R^2={:.2f}$".format(r**2), transform=ax3.transAxes)

    if heteroscedasicity and sim is not None:
        # Plot residuals vs. simulation
        # interpolate simulation to times of observations
        sim = sim.loc[series.index]
        ax4.plot(sim, series, marker=".", linestyle=" ", color=c, alpha=0.7)
        ax4.grid()
        ax4.set_xlabel("Simulated values")
        ax4.set_ylabel("Residuals")

        # Plot residuals vs. simulation
        ax5.plot(
            sim, np.sqrt(series.abs()), marker=".", linestyle=" ", color=c, alpha=0.7
        )
        ax5.set_xlabel("Simulated values")
        ax5.set_ylabel("$\\sqrt{|Residuals|}$")
        ax5.grid()

    return fig.axes


def cum_frequency(
    obs: Series,
    sim: Series | None = None,
    ax: Axes | None = None,
    figsize: tuple = (5, 2),
) -> Axes:
    """Plot of the cumulative frequency of a time Series.

    Parameters
    ----------
    sim: pandas.Series
        Series with the simulated values.
    obs: pandas.Series
        The pandas Series with the observed values.
    ax: matplotlib.axes.Axes, optional
        Matplotlib Axes instance to create the plot on. A new Figure and Axes is
        created when no value for ax is provided.
    figsize: tuple, optional
        2-D Tuple to determine the size of the figure created. Ignored if ax is also
        provided.

    Returns
    -------
    ax: matplotlib.axes.Axes

    Examples
    --------
    >>> obs = pd.Series(index=pd.date_range(start=0, periods=1000, freq="D"),
    >>>                 data=np.random.normal(0, 1, 1000))
    >>> ps.stats.plot_cum_frequency(obs)
    """
    if ax is None:
        _, ax = plt.subplots(1, 1, figsize=figsize)

    ax.plot(
        obs.sort_values(),
        np.arange(0, obs.size) / obs.size * 100,
        color="k",
        marker=".",
        linestyle=" ",
    )
    if sim is not None:
        ax.plot(sim.sort_values(), np.arange(0, sim.size) / sim.size * 100)
    ax.legend(["Observations", "Simulation"])
    ax.set_xlabel("Head")
    ax.set_ylabel("Cum. Frequency [%]")
    ax.grid()
    plt.tight_layout()

    return ax


class TrackSolve:
    """Track and/or visualize optimization progress for Pastas models.

    Parameters
    ----------
    ml : pastas.model.Model
        pastas Model to track
    tmin : str or pandas.Timestamp, optional
        start time for simulation, by default None which defaults to first index in
        ml.oseries.series
    tmax : str or pandas.Timestamp, optional
        end time for simulation, by default None which defaults to last index in
        ml.oseries.series
    update_iter : int, optional
        if visualizing optimization progress, update plot every update_iter
        iterations, by default nparam

    Notes
    -----
    Interactive plotting of optimization progress requires a matplotlib backend that
    supports interactive plotting, e.g. `mpl.use("TkAgg")` and `mpl.interactive(
    True)`. Some possible speedups on the matplotlib side include:
    - mpl.style.use("fast")
    - mpl.rcParams['path.simplify_threshold'] = 1.0

    Examples
    --------
    Set matplotlib backend and interactive mode (put this at the top of your script)::

        import matplotlib as mpl
        mpl.use("TkAgg")
        import matplotlib.pyplot as plt
        plt.ion()

    Create a TrackSolve object for your model::

        track = TrackSolve(ml)

    Solve model and store intermediate optimization results::

        ml.solve(callback=track.track_solve)

    Calculated parameters per iteration are stored in a pandas.DataFrame::

        track.parameters

    Other stored statistics include `track.evp` (explained variance percentage),
    `track.rmse_res` (root-mean-squared error of the residuals), `track.rmse_noise` (
    root mean squared error of the noise, only if noise=True).

    To interactively plot model optimization progress while solving pass
    `track.plot_track_solve` as callback function::

        ml.solve(callback=track.plot_track_solve)

    Access the resulting figure through `track.fig`.
    """

    def __init__(
        self,
        ml: Model,
        tmin: TimestampType | None = None,
        tmax: TimestampType | None = None,
        update_iter: int | None = None,
    ) -> None:
        logger.warning(
            "TrackSolve feature under development. If you find any bugs please post "
            "an issue on GitHub: https://github.com/pastas/pastas/issues"
        )

        self.ml = ml
        self.viewlim = 75  # no of iterations on axes by default
        if update_iter is None:
            self.update_iter = len(
                self.ml.parameters.loc[self.ml.parameters.vary].index
            )
        else:
            self.update_iter = update_iter  # update plot every update_iter

        # get tmin/tmax
        if tmin is None:
            self.tmin = self.ml.oseries.series.index[0]
        else:
            self.tmin = Timestamp(tmin)

        if tmax is None:
            self.tmax = self.ml.oseries.series.index[-1]
        else:
            self.tmax = Timestamp(tmax)

        # parameters
        self.parameters = DataFrame(columns=self.ml.parameters.index)
        self.parameters.loc[0] = self.ml.parameters.initial.values

        # iteration counter
        self.itercount = 0

        # calculate RMSE residuals
        res = self._residuals(self.ml.parameters.initial.values)
        self.rmse_res = np.array([rmse(res=res)])

        # calculate RMSE noise
        if self.ml.settings["noise"] and self.ml.noisemodel is not None:
            noise = self._noise(self.ml.parameters.initial.values)
            self.rmse_noise = np.array([rmse(res=noise)])

        # get observations
        self.obs = self.ml.observations(tmin=self.tmin, tmax=self.tmax)
        # calculate EVP
        self.evp = np.array([evp(obs=self.obs, res=res)])

    def track_solve(self, params: ArrayLike) -> None:
        """Append parameters to self.parameters DataFrame and update itercount,
        rmse values and evp.

        Parameters
        ----------
        params : array_like
            array containing parameters.
        """
        # update tmin/tmax and freq once after starting solve
        if self.itercount == 0:
            self._update_settings()

        # update itercount
        self.itercount += 1

        # add parameters to DataFrame
        self.parameters.loc[self.itercount, self.ml.parameters.index] = params.copy()

        # calculate new RMSE values
        r_res = self._residuals(params)
        self.rmse_res = np.r_[self.rmse_res, rmse(res=r_res)]

        if self.ml.settings["noise"] and self.ml.noisemodel is not None:
            n_res = self._noise(params)
            self.rmse_noise = np.r_[self.rmse_noise, rmse(res=n_res)]

        # recalculate EVP
        self.evp = np.r_[self.evp, evp(obs=self.obs, res=r_res)]

    def _update_axes(self) -> None:
        """extend xlim if number of iterations exceeds current window."""
        for iax in self.axes[1:]:
            iax.set_xlim(right=self.viewlim)
            self.fig.canvas.draw()

    def _update_settings(self) -> None:
        self.tmin = self.ml.settings["tmin"]
        self.tmax = self.ml.settings["tmax"]
        self.freq = self.ml.settings["freq"]

    def _noise(self, params: ArrayLike) -> ArrayLike:
        """get noise.

        Parameters
        ----------
        params: array_like
            array containing parameters.

        Returns
        -------
        noise: array_like
            array containing noise.
        """
        noise = self.ml.noise(p=params, tmin=self.tmin, tmax=self.tmax)
        return noise

    def _residuals(self, params: ArrayLike) -> ArrayLike:
        """calculate residuals.

        Parameters
        ----------
        params: np.array
            array containing parameters.

        Returns
        -------
        res: array_like
            array containing residuals.
        """
        res = self.ml.residuals(p=params, tmin=self.tmin, tmax=self.tmax)
        return res

    def _simulate(self) -> Series:
        """simulate model with last entry in self.parameters.

        Returns
        -------
        sim: pd.Series
            Series containing model evaluation.
        """
        sim = self.ml.simulate(
            p=self.parameters.iloc[-1, :].values,
            tmin=self.tmin,
            tmax=self.tmax,
            freq=self.ml.settings["freq"],
        )
        return sim

    def initialize_figure(
        self, figsize: tuple[int] = (10, 8), dpi: int = 100
    ) -> Figure:
        """Initialize figure for plotting optimization progress.

        Parameters
        ----------
        figsize : tuple, optional
            figure size, passed to plt.subplots(), by default (10, 8).
        dpi : int, optional
            dpi of the figure passed to plt.subplots(), by default 100.

        Returns
        -------
        fig : matplotlib.pyplot.Figure
            handle to the figure.
        """
        # create plot
        self.fig, self.axes = plt.subplots(3, 1, figsize=figsize, dpi=dpi)
        self.ax0, self.ax1, self.ax2 = self.axes

        # share x-axes between 2nd and 3rd axes
        self.ax1.sharex(self.ax2)
        for t in self.ax1.get_xticklabels():
            t.set_visible(False)

        # plot oseries
        self.ax0.plot(
            self.obs.index,
            self.obs,
            marker=".",
            ls="none",
            label="observations",
            color="k",
            ms=4,
        )

        # plot simulation
        sim = self._simulate()
        (self.simplot,) = self.ax0.plot(sim.index, sim, label="simulation")
        self.ax0.set_ylabel("head")
        self.ax0.set_title(
            "Iteration: {0} (EVP: {1:.2f}%)".format(self.itercount, self.evp[-1])
        )
        self.ax0.legend(loc=(0, 1), frameon=False, ncol=2)
        omax = self.obs.max()
        omin = self.obs.min()
        vspace = 0.05 * (omax - omin)
        self.ax0.set_ylim(bottom=omin - vspace, top=omax + vspace)

        # plot RMSE (residuals and/or residuals)
        plt.sca(self.ax1)
        plt.yscale("log")
        legend_handles = []
        (self.r_rmse_plot_line,) = self.ax1.plot(
            [0], self.rmse_res[0:1], c="k", ls="solid", label="residuals"
        )
        (self.r_rmse_plot_dot,) = self.ax1.plot(
            self.itercount, self.rmse_res[-1], c="k", marker="o", ls="none"
        )
        legend_handles.append(self.r_rmse_plot_line)
        self.ax1.set_xlim(0, self.viewlim)
        self.ax1.set_ylim(1e-2, 2 * self.rmse_res[-1])
        self.ax1.set_ylabel("RMSE")

        if self.ml.settings["noise"] and self.ml.noisemodel is not None:
            (self.n_rmse_plot_line,) = self.ax1.plot(
                [0], self.rmse_noise[0:1], c="C0", ls="solid", label="noise"
            )
            (self.n_rmse_plot_dot,) = self.ax1.plot(
                self.itercount, self.rmse_res[-1], c="C0", marker="o", ls="none"
            )
            legend_handles.append(self.n_rmse_plot_line)
        legend_labels = [i.get_label() for i in legend_handles]
        self.ax1.legend(
            legend_handles, legend_labels, loc=(0, 1), frameon=False, ncol=2
        )

        # plot parameters values on semilogy
        plt.sca(self.ax2)
        plt.yscale("log")
        self.param_plot_handles = []
        legend_handles = []
        for pname, row in self.ml.parameters.iterrows():
            if pname.startswith("noise"):
                if not self.ml.settings["noise"] or self.ml.noisemodel is None:
                    continue
            (pa,) = self.ax2.plot(
                [0], np.abs(row.initial), marker=".", ls="none", label=pname
            )
            (pb,) = self.ax2.plot(
                [0], np.abs(row.initial), ls="solid", c=pa.get_color()
            )
            self.param_plot_handles.append((pa, pb))
            legend_handles.append(pa)

        legend_labels = [i.get_label() for i in legend_handles]
        self.ax2.legend(
            legend_handles, legend_labels, loc=(0, 1), ncol=6, frameon=False
        )
        self.ax2.set_xlim(0, self.viewlim)
        self.ax2.set_ylim(1e-3, 1e4)
        self.ax2.set_ylabel("Parameter values")
        self.ax2.set_xlabel("Iteration")

        # set grid for each plot
        for iax in [self.ax0, self.ax1, self.ax2]:
            iax.grid(visible=True)

        self.fig.align_ylabels()
        self.fig.tight_layout()
        return self.fig

    def plot_track_solve(self, params: ArrayLike) -> None:
        """Method to plot model simulation while model is being solved.

        Parameters
        ----------
        params : array_like
            array containing parameters

        Examples
        --------
        Pass
        this method to ml.solve(), e.g.:

        >>> track = TrackSolve(ml)
        >>> ml.solve(callback=track.plot_track_solve)

        """
        if not hasattr(self, "fig"):
            self.initialize_figure()

        # update parameters
        self.track_solve(params)

        # check if figure should be updated
        if self.itercount % self.update_iter != 0:
            return

        # update view limits if needed
        if self.itercount >= self.viewlim:
            self.viewlim += 50
            self._update_axes()

        # update simulation
        sim = self._simulate()
        self.simplot.set_data(sim.index, sim.values)

        # update rmse residuals
        self.r_rmse_plot_line.set_data(
            range(self.itercount + 1), np.array(self.rmse_res)
        )
        self.r_rmse_plot_dot.set_data(
            np.array([self.itercount]), np.array([self.rmse_res[-1]])
        )

        if self.ml.settings["noise"] and self.ml.noisemodel is not None:
            # update rmse noise
            self.n_rmse_plot_line.set_data(
                range(self.itercount + 1), np.array(self.rmse_noise)
            )
            self.n_rmse_plot_dot.set_data(
                np.array([self.itercount]), np.array([self.rmse_noise[-1]])
            )

        # update parameter plots
        for j, (p1, p2) in enumerate(self.param_plot_handles):
            p1.set_data(
                np.array([self.itercount]), np.abs([self.parameters.iloc[-1, j]])
            )
            p2.set_data(
                range(self.itercount + 1), self.parameters.iloc[:, j].abs().values
            )

        # update title
        self.ax0.set_title(
            "Iteration: {0} (EVP: {1:.2f}%)".format(self.itercount, self.evp[-1])
        )
        plt.pause(1e-10)
        self.fig.canvas.draw()

    def plot_track_solve_history(self, fig: Figure | None = None) -> list[Axes]:
        """Plot optimization history.

        Parameters
        ----------
        fig : matplotlib.pyplot.Figure, optional
            figure handle, by default None, which constructs a new figure with
            `self.initialize_figure()`.

        Returns
        -------
        axes : list of matplotlib.pyplot.Axes
            list of axes handles in figure.
        """

        if fig is None:
            fig = self.initialize_figure()
        self.plot_track_solve(self.ml.parameters.optimal.values)

        self.fig.axes[1].autoscale(tight=False, axis="both")
        self.fig.axes[2].autoscale(tight=False, axis="both")

        self.fig.axes[1].set_xlim(left=0)
        # because of bug with autoscaling log axis?
        self.fig.axes[1].set_ylim(top=1.05 * self.rmse_res.max())

        return fig.axes


def pairplot(
    data: DataFrame | list[Series],
    bins: int | None = None,
) -> dict[str, Axes]:
    """Plot correlation between time series on of values on the same time steps.

    Based on seaborn pairplot method.

    Parameters
    ----------
    data : DataFrame | list[Series]
        List of Series or Dataframe with DateTime index
    bins : int | None, optional
        Number of bins in the histogram, by default None which uses Sturge's
        Rule to determine the number bins

    Returns
    -------
    dict[str, Axes]
    """
    if isinstance(data, list):
        data = concat(data, axis=1)

    df = data.dropna(how="any")

    columns = df.columns

    mosaic = []
    for i, column in enumerate(columns):
        cols = [f"scatter_{x}-{column}" for x in columns]
        cols[i] = f"hist_{column}"
        mosaic.append(cols)

    mosaic = np.array(mosaic)

    f, axd = plt.subplot_mosaic(mosaic, figsize=(6.5, 6))

    for i, (column, mos) in enumerate(zip(columns, mosaic)):
        # plot histogram
        if bins is None:
            bins = int(np.ceil(1 + np.log2(len(df.loc[:, column].values))))
        counts, bins = np.histogram(df.loc[:, column].values, bins=bins)
        scaled_counts = (
            df.loc[:, column].max()
            * (counts - np.min(counts))
            / (np.max(counts) - np.min(counts))
        )
        axd[f"hist_{column}"].hist(x=bins[:-1], bins=bins, weights=scaled_counts)

        # plot scatter
        other_cols = [x for x in columns if x is not column]
        for col in other_cols:
            axd[f"scatter_{column}-{col}"].scatter(
                df.loc[:, column].values,
                df.loc[:, col].values,
                alpha=0.6,
                s=20,
                edgecolor="white",
                linewidth=0.3,
            )
            r, _ = pearsonr(df.loc[:, column].values, df.loc[:, col].values)
            axd[f"scatter_{column}-{col}"].annotate(
                f"r = {r:.2f}",
                xy=(0.5, 0.95),
                horizontalalignment="center",
                verticalalignment="top",
                xycoords="axes fraction",
                color="k",
                path_effects=[
                    path_effects.withStroke(linewidth=2, foreground="white"),
                    path_effects.Normal(),
                ],
            )

        # set labels
        axd[mos[0]].set_ylabel(column)
        if (mos == mosaic[-1]).all():
            _ = [axd[j].set_xlabel(x) for x, j in zip(columns, mos)]

        # share x and y axis per row and columns
        share_yaxes([axd[j] for j in mos])

    _ = [share_xaxes([axd[x] for x in mosaic[:, j]]) for j in range(len(columns))]

    f.tight_layout()

    return axd



"""This file contains the import method for pas-files.

Import a .pas file (basically a json format)

R.A. Collenteur - August 2017
"""

import datetime
import json
from collections import OrderedDict
from io import StringIO as stringIO
from logging import getLogger

from numpy import integer

try:
    from shapely.geometry.base import BaseGeometry

    SHAPELY = True
except ModuleNotFoundError:
    SHAPELY = False
    BaseGeometry = None

from pandas import (
    DataFrame,
    Series,
    Timedelta,
    Timestamp,
    isna,
    read_json,
    to_timedelta,
)

logger = getLogger(__name__)


def load(fname: str) -> dict:
    with open(fname, "r") as file:
        data = json.load(file, object_hook=pastas_hook)
    return data


def pastas_hook(obj: dict):
    for key, value in obj.items():
        if key in ["tmin", "tmax", "date_modified", "date_created"]:
            val = Timestamp(value)
            if isna(val):
                val = None
            obj[key] = val
        elif key == "series":
            try:
                obj[key] = read_json(stringIO(value), typ="series", orient="split")
            except Exception as e:
                logger.debug(e)
                obj[key] = value
            if isinstance(obj[key], Series):
                obj[key].index = obj[key].index.tz_localize(None)
        elif key in ["time_offset", "warmup"]:
            if isinstance(value, int) or isinstance(value, float):
                obj[key] = Timedelta(value, "d")
            else:
                obj[key] = Timedelta(value)
        elif key in ["parameters", "pcov"]:
            if value is not None:
                # Necessary to maintain order when using the JSON format!
                value = json.loads(value, object_pairs_hook=OrderedDict)
                param = DataFrame(data=value, columns=value.keys()).T
                obj[key] = param.infer_objects()
            else:
                obj[key] = value
        else:
            try:
                obj[key] = json.loads(value, object_hook=pastas_hook)
            except Exception as e:
                logger.debug(e)
                obj[key] = value
    return obj


def dump(fname: str, data: dict) -> None:
    with open(fname, "w") as file:
        json.dump(data, file, indent=4, cls=PastasEncoder)
    logger.info("%s file successfully exported", fname)


class PastasEncoder(json.JSONEncoder):
    """Enhanced encoder to deal with the pandas formats used throughout Pastas.

    Notes
    -----
    Currently supported formats are: DataFrame, Series, Timedelta, Timestamps.

    see: https://docs.python.org/3/library/json.html
    """

    def default(self, o):
        if isinstance(o, (Timestamp, datetime.datetime, datetime.date)):
            return o.isoformat()
        elif isinstance(o, Series):
            return o.to_json(date_format="iso", orient="split")
        elif isinstance(o, DataFrame):
            # Necessary to maintain order when using the JSON format!
            # Do not use o.to_json() because of float precision
            return json.dumps(o.to_dict(orient="index"), indent=0)
        elif isinstance(o, (Timedelta, datetime.timedelta)):
            if isinstance(o, datetime.timedelta):
                o = to_timedelta(o)
            return o.to_timedelta64().__str__()
        elif SHAPELY and isinstance(o, BaseGeometry):
            return o.wkt
        elif isna(o):
            return None
        elif isinstance(o, integer):
            return int(o)
        else:
            return super(PastasEncoder, self).default(o)



from collections.abc import Callable
from typing import Any

import pandas as pd
import pytest

from pastas import Model, check


def test_response_memory(ml_noisemodel: Model) -> None:
    df = check.response_memory(ml_noisemodel, cutoff=0.95, factor_length_oseries=0.5)
    assert df["pass"].item()


def test_rsq_geq_threshold(ml_noisemodel: Model) -> None:
    df = check.rsq_geq_threshold(ml_noisemodel, threshold=0.7)
    assert df["pass"].item()


def test_acf_runs_test(ml_noisemodel: Model) -> None:
    df = check.acf_runs_test(ml_noisemodel)
    assert df["pass"].item()


def test_acf_stoffer_toloi(ml_noisemodel: Model) -> None:
    df = check.acf_stoffer_toloi_test(ml_noisemodel)
    assert df["pass"].item()


def test_parameter_bounds(ml_noisemodel: Model) -> None:
    df = check.parameter_bounds(ml_noisemodel)
    assert df["pass"].all().item()


def test_parameter_uncertainty(ml_noisemodel: Model) -> None:
    df = check.uncertainty_parameters(ml_noisemodel, n_std=2)
    assert df["pass"].all().item()


def test_uncertainty_gain(ml_noisemodel: Model) -> None:
    df = check.uncertainty_gain(ml_noisemodel, n_std=2)
    assert df["pass"].item()


def test_checklist(ml_noisemodel: Model) -> None:
    df = check.checklist(ml_noisemodel, check.checks_brakenhoff_2022)
    assert df["pass"].all().item()


# New tests to improve test coverage


class TestOperatorsAndUtilities:
    """Test the operator mapping and utility functions."""

    def test_operators_dict(self) -> None:
        """Test that the operators dictionary has correct mappings."""
        assert check.operators["greater_equal"] == ">="
        assert check.operators["less_equal"] == "<="  # Check the fixed operator
        assert check.operators["greater_than"] == ">"
        assert check.operators["less_than"] == "<"
        assert check.operators["equal"] == "=="
        assert check.operators["not_equal"] != "=="

    def test_guess_unit_or_dims(self) -> None:
        """Test the guess_unit_or_dims function."""
        # Test _A parameter
        result = check.guess_unit_or_dims("precip_A")
        assert "[L]" in result
        assert "precip" in result

        # Test noise_alpha
        result = check.guess_unit_or_dims("noise_alpha", return_dims=False)
        assert result == "days"

        # Test constant_d
        result = check.guess_unit_or_dims("constant_d")
        assert result == "[L]"

        # Test _f parameter
        result = check.guess_unit_or_dims("precip_f")
        assert result == "[-]"

        # Test unknown parameter
        result = check.guess_unit_or_dims("unknown_parameter")
        assert result == ""


class TestChecklistFunction:
    """Test the checklist function with different input types."""

    def test_checklist_with_string(self, ml_noisemodel: Model) -> None:
        """Test checklist with string function name."""
        checks: list[str] = ["rsq_geq_threshold"]
        result = check.checklist(ml_noisemodel, checks, report=False)

        # Check that result has the expected format
        assert isinstance(result, pd.DataFrame)
        assert "rsq>=0.7" in result.index

    def test_checklist_with_callable(self, ml_noisemodel: Model) -> None:
        """Test checklist with callable function."""
        checks: list[Callable[[Model], pd.DataFrame]] = [
            lambda ml: check.rsq_geq_threshold(ml, threshold=0.8)
        ]
        result = check.checklist(ml_noisemodel, checks, report=False)

        # Check that result has the expected format
        assert isinstance(result, pd.DataFrame)
        assert "rsq>=0.8" in result.index

    def test_checklist_with_dict(self, ml_noisemodel: Model) -> None:
        """Test checklist with dictionary."""
        checks: list[dict[str, str | float]] = [
            {"func": "rsq_geq_threshold", "threshold": 0.9}
        ]
        result = check.checklist(ml_noisemodel, checks, report=False)

        # Check that result has the expected format
        assert isinstance(result, pd.DataFrame)
        assert "rsq>=0.9" in result.index

    def test_checklist_with_invalid_type(self, ml_noisemodel: Model) -> None:
        """Test checklist with invalid type."""
        checks: list[int] = [123]  # Not a string, callable, or dict
        with pytest.raises(TypeError):
            check.checklist(ml_noisemodel, checks, report=False)


@pytest.mark.parametrize(
    "check_func,kwargs",
    [
        (check.rsq_geq_threshold, {"threshold": 0.5}),
        (check.acf_runs_test, {"p_threshold": 0.01}),
        (check.acf_stoffer_toloi_test, {"p_threshold": 0.01}),
    ],
)
def test_check_functions_parameterized(
    ml_noisemodel: Model,
    check_func: Callable[[Model, Any], pd.DataFrame],
    kwargs: dict[str, float],
) -> None:
    """Test various check functions with different parameters."""
    df = check_func(ml_noisemodel, **kwargs)
    assert isinstance(df, pd.DataFrame)
    assert "pass" in df.columns
    assert "statistic" in df.columns
    assert "threshold" in df.columns



"""Tests for numba-accelerated functions in stressmodels.py."""

import numpy as np

from pastas.stressmodels import TarsoModel


class TestTarsoModelNumba:
    """Test the numba-accelerated tarso function in TarsoModel."""

    def setup_method(self) -> None:
        """Setup for tests."""
        # Create test data for recharge input
        np.random.seed(42)  # For reproducibility
        self.npoints = 100
        self.r = np.random.normal(1, 0.5, self.npoints)  # Simple recharge series

    def test_tarso_py_func(self) -> None:
        """Test the Python implementation of the tarso function."""
        # Access the Python version directly via py_func attribute
        tarso_py_func = TarsoModel.tarso.py_func

        # Parameters (A0, a0, d0, A1, a1, d1)
        p = np.array([0.9, 10.0, 4.0, 0.8, 15.0, 6.0])
        dt = 1.0

        # Call the Python function directly
        result = tarso_py_func(p, self.r, dt)

        # Check basic properties
        assert len(result) == len(self.r)
        assert not np.isnan(result).any()

        # Result should be bounded by drainage levels plus some margin
        assert np.all(result >= p[2] - 1)  # Near or above d0
        assert np.all(result <= p[5] + 1)  # Near or below d1

    def test_tarso_parameter_variations(self) -> None:
        """Test tarso function with different parameter values."""
        tarso_py_func = TarsoModel.tarso.py_func
        dt = 1.0

        # Test different parameter combinations
        test_parameters = [
            # Different drainage level scenarios
            np.array([0.8, 10.0, 2.0, 0.7, 12.0, 4.0]),  # Lower drainage levels
            np.array([0.9, 15.0, 6.0, 0.8, 20.0, 8.0]),  # Higher drainage levels
            np.array([0.9, 10.0, 4.0, 0.8, 15.0, 4.5]),  # Close drainage levels
            # Different response characteristics
            np.array([0.5, 5.0, 4.0, 0.4, 7.0, 6.0]),  # Faster response
            np.array([1.5, 25.0, 4.0, 1.3, 30.0, 6.0]),  # Slower response
        ]

        for p in test_parameters:
            result = tarso_py_func(p, self.r, dt)

            # Basic checks for each parameter set
            assert len(result) == len(self.r)
            assert not np.isnan(result).any()

            # Results should generally be bounded by drainage levels
            assert np.all(result >= p[2] - 2)
            assert np.all(result <= p[5] + 2)

    def test_tarso_dt_variations(self) -> None:
        """Test tarso function with different time step values."""
        tarso_py_func = TarsoModel.tarso.py_func
        p = np.array([0.9, 10.0, 4.0, 0.8, 15.0, 6.0])

        # Test different dt values
        for dt in [0.1, 0.5, 1.0, 2.0, 5.0]:
            result = tarso_py_func(p, self.r, dt)

            # Basic checks
            assert len(result) == len(self.r)
            assert not np.isnan(result).any()



import pytest
from pandas import Series, concat

import pastas as ps

tol = 1e-3


def test_sgi_period1(head: Series) -> None:
    sgi = ps.stats.sgi(head, timescale_months=1)
    assert pytest.approx(sgi.iat[30], tol) == -0.812


def test_sgi_df(head: Series) -> None:
    df = concat([head, head], axis=1)
    sgi = ps.stats.sgi(df, timescale_months=1)
    assert pytest.approx(sgi.iat[30, 0], tol) == -0.812
    assert pytest.approx(sgi.iat[30, 1], tol) == -0.812


def test_sgi_period2(head: Series) -> None:
    sgi = ps.stats.sgi(head, timescale_months=2)
    assert pytest.approx(sgi.iat[30], tol) == -1.258


def test_sgi_perioderror(head: Series) -> None:
    with pytest.raises(ValueError):
        _ = ps.stats.sgi(head, timescale_months=4)



import os
from pathlib import Path

import matplotlib.pyplot as plt
import pytest

pathname = Path(__file__).parent.parent / "doc/examples"
files = list(pathname.glob("*.py"))


@pytest.mark.parametrize("file", files)
def test_example(file) -> None:
    cwd = os.getcwd()
    os.chdir(pathname)
    try:
        # run each example
        with open(file) as f:
            exec(compile(f.read(), file, "exec"))
            # Report success
            print(f"Example {file} ran successfully.")
        plt.close("all")
    except Exception as e:
        os.chdir(cwd)
        raise Exception(f"could not run {file}") from e
    os.chdir(cwd)


if __name__ == "__main__":
    for file in files:
        test_example(file)



"""This module contains methods to generate forecasts using a Pastas model instance."""

from logging import getLogger

from numpy import array, empty, exp, linspace, ones
from pandas import DataFrame, DatetimeIndex, MultiIndex, Timedelta, concat

from pastas.model import Model
from pastas.noisemodels import ArNoiseModel
from pastas.typing import ArrayLike, TimestampType

logger = getLogger(__name__)


def _check_forecast_data(
    forecasts: dict[str, list[DataFrame]],
) -> tuple[int, TimestampType, TimestampType, DatetimeIndex]:
    """Internal method to check the integrity of the forecasts data.

    Parameters
    ----------
    forecasts: dict
        Dictionary containing the forecasts data. The keys are the stressmodel names
        and the values are lists of DataFrames containing the forecasts with a datetime
        index and each column a time series (i.e., one ensemble member).

    Returns
    -------
    n: int
        The number of ensemble members in the forecasts.
    tmin: datetime
        The minimum datetime in the forecasts.
    tmax: datetime
        The maximum datetime in the forecasts.
    index: DatetimeIndex
        The datetime index of the forecasts.

    Notes
    -----
    This method checks if the number of columns and indices are the same for all
    DataFrames in the forecasts dictionary. If the number of columns or the indices are
    not the same, a warning is printed and a ValueError is raised.

    """
    # Input validation
    if not isinstance(forecasts, dict) or not forecasts:
        msg = "Forecasts must be a non-empty dictionary"
        logger.error(msg)
        raise ValueError(msg)

    n = None
    tmax = None
    tmin = None
    index = None

    for sm_name, fc_data in forecasts.items():
        if not fc_data:
            msg = f"No forecast data provided for stressmodel '{sm_name}'"
            logger.warning(msg)
            continue

        for fc in fc_data:
            # Check if DataFrame is empty
            if fc.empty:
                msg = f"Empty DataFrame in forecasts for stressmodel '{sm_name}'"
                logger.warning(msg)
                continue

            # Check if the number of columns is the same for all DataFrames
            if n is None:
                n = fc.columns.size
                tmin = fc.index[0]
                tmax = fc.index[-1]
                index = fc.index
                logger.debug(f"First forecast found with {n} ensemble members")
            # If the number of columns is not the same, raise an error
            elif n != fc.columns.size:
                msg = (
                    f"The number of ensemble members is not the same for all forecasts. "
                    f"Expected {n}, got {fc.columns.size} in stressmodel '{sm_name}'."
                )
                logger.error(msg)
                raise ValueError(msg)
            elif tmin != fc.index[0] or tmax != fc.index[-1]:
                msg = (
                    "The time index of the forecasts is not the same for all forecasts."
                    "Please check the forecast data."
                )
                logger.error(msg)
                raise ValueError(msg)

    if n is None:
        msg = "No valid forecast data found in any of the stressmodels"
        logger.error(msg)
        raise ValueError(msg)

    return n, tmin, tmax, index


def forecast(
    ml: Model,
    forecasts: dict[str, list[DataFrame]],
    p: ArrayLike | None = None,
    post_process: bool = False,
) -> DataFrame:
    """Method to forecast the head from ensembles of stress forecasts.

    Parameters
    ----------
    ml: pastas.Model
        Pastas Model instance.
    forecasts: dict
        Dictionary containing the forecasts data. The keys are the stressmodel names
        and the values are lists of DataFrames containing the forecasts with a datetime
        index and each column a time series (i.e., one ensemble member).
    p: array_like, optional
        List of parameter sets to use for the forecasts. If None, a single parameter set is used that defaults to the optimal model parameters. Default is None.
    post_process: bool, optional
        If True, the forecasts are post-processed using the noise model of the model
        instance. Default is False. If True, a noise model should be present in the
        model instance. If no noisemodel is present and post_process is True, an error
        is raised.

    Returns
    -------
    df: pandas.DataFrame
        DataFrame containing the forecasts. The columns are a MultiIndex with the first
        level the ensemble member, the second level the parameter member, and the third level the mean and the variance of each forecast member.

    Notes
    -----
    For efficiency, the iteration over the different ensemble members and parameters sets is done in the following order using a double for-loop:

    1. iterate over the ensemble members
    2. iterate over the parameter sets

    All computations that only depend on the parameter sets and not on the ensemble members are performed outside this double for-loop.

    Please note that only the AR1 noise model is supported at this moment for post-processing.

    """
    # Check the integrity of the forecasts data
    n, tmin, tmax, index = _check_forecast_data(forecasts)
    logger.info(f"Working with {n} ensemble members from {tmin} to {tmax}")

    if post_process and not isinstance(ml.noisemodel, ArNoiseModel):
        if ml.noisemodel is None:
            msg = "No noise model present in the model instance. Please add a noise model to the model instance or set post_process=False."
        else:
            msg = "Only the AR1 noise model is supported for post-processing at this moment. Please use an AR1 noise model or set post_process=False."
        logger.error(msg)
        raise ValueError(msg)

    # Check which parameters sets should be used.
    if p is None:
        logger.info("No parameter provided, using the optimal parameters.")
        # In case no parameters are provided, use optimal values
        p = [ml.parameters.loc[:, "optimal"].values]
        nparam = len(p)
    else:
        if len(p) == 0:
            msg = "Empty parameter list provided"
            logger.error(msg)
            raise ValueError(msg)
        nparam = len(p)
        logger.info(f"Using {nparam} provided parameter sets")

    # Pre-allocate arrays for results to avoid append operations
    forecast_length = len(index)
    result_array = empty((n * nparam * 2, forecast_length))

    # Pre-compute residuals for each parameter set since they only depend on parameters
    logger.info("Pre-computing residuals for each parameter set")

    residuals = {}
    vars = {}
    day = Timedelta("1D")

    if post_process:
        dt = ml.settings["freq_obs"] / day
        t = linspace(1, index.size, index.size)
        correction = {}

    # Preprocess residuals and variances for each parameter set as they only depend on parameters and not on ensemble members
    for i, param in enumerate(p):
        residuals[i] = ml.residuals(tmax=tmin, p=param).dropna()

        if post_process:
            # Compute the time varying variance for the AR1 noise model
            phi = exp(-dt / param[-1])
            denominator = 1.0 - phi**2
            phi_scaling_factor = (1.0 - phi ** (2.0 * t / dt)) / denominator
            vars[i] = ml.noise(tmax=tmin, p=param).var() * phi_scaling_factor

            correction[i] = ml.noisemodel.get_correction(
                residuals[i], [param[-1]], index
            ).values
        else:
            vars[i] = residuals[i].var() * ones(forecast_length)

    # Copy the model so old model is unaffected when replacing the stresses.
    ml = ml.copy()
    idx = 0

    # 1. iterate over the ensemble members
    for member in range(n):
        # Update stresses with ensemble member data
        for sm_name, fc_data in forecasts.items():
            sm = ml.stressmodels[sm_name]  # Select stressmodel
            for i, fc in enumerate(fc_data):
                ts = concat(
                    [
                        sm.stress[i].series_original.loc[: tmin - day],
                        fc.iloc[:, member],
                    ]
                )
                sm.stress[i].series_original = ts

        # 2. iterate over the parameter sets
        for i, param in enumerate(p):
            # Generate the forecasts
            sim = ml.simulate(tmin=tmin, tmax=tmax, p=param).values

            if post_process:
                # Add the correction from the noise model
                sim = sim + correction[i]

            # Store in pre-allocated array instead of appending to list
            result_array[idx : idx + 2] = array([sim, vars[i]])
            idx += 2  # Bump index by 2 for mean and variance

    # Create DataFrames to store data
    mi = MultiIndex.from_product(
        [range(n), range(nparam), ["mean", "var"]],
        names=["ensemble_member", "param_member", "forecast"],
    )
    df = DataFrame(data=result_array.T, index=index, columns=mi, dtype=float)

    return df


def get_overall_mean_and_variance(df: DataFrame) -> tuple[DataFrame, DataFrame]:
    """Method to get the overall mean and variance of the forecast ensemble.

    Parameters
    ----------
    df: pandas.DataFrame
        DataFrame containing the forecasts. The columns are a MultiIndex with the first
        level the ensemble member, the second level the parameter member, and the third
        level the mean and the variance of each forecast member. The index is a
        DatetimeIndex with the time steps of the forecasts.

    Returns
    -------
    overall_mean: pandas.Series
        Series with the mean of the forecasts.
    overall_variance: pandas.Series
        Series with the variance of the forecasts.

    Notes
    -----
    This method is used to get the overall mean and variance of the forecasts. The
    mean and variance are calculated from the ensemble members and parameter members
    using the law of total variance.

    Example
    -------
    >>> import pastas as ps
    >>> import pandas as pd
    >>> import numpy as np
    >>> from pastas.forecast import get_overall_mean_and_variance

    >>> # Create a sample DataFrame with forecasts
    >>> index = pd.date_range("2023-01-01", periods=10, freq="D")
    >>> data = np.random.rand(10, 6)
    >>> columns = pd.MultiIndex.from_product(
    ...     [range(3), range(2), ["mean", "var"]],
    ...     names=["ensemble_member", "param_member", "forecast"],
    ... )
    >>> df = pd.DataFrame(data=data, index=index, columns=columns)
    >>> # Call the function to get the overall mean and variance
    >>> mean, var = get_overall_mean_and_variance(df)
    >>> print(mean)

    """
    means = df.loc[:, (slice(None), slice(None), "mean")]
    variances = df.loc[:, (slice(None), slice(None), "var")]

    overall_mean = means.mean(axis=1)

    # variance of the means
    variance_of_means = means.var(axis=1)

    # mean of the variances
    mean_of_variances = variances.mean(axis=1)

    # total variance
    overall_variance = variance_of_means + mean_of_variances

    return overall_mean, overall_variance



"""Tests for ensuring code coverage of Numba-accelerated functions in stats.core."""

import numpy as np
from numpy.random import random
from numpy.testing import assert_array_almost_equal

from pastas.stats.core import _compute_ccf_gaussian, _compute_ccf_rectangle


def test_compute_ccf_rectangle_py_func() -> None:
    """Test the Python version of _compute_ccf_rectangle for coverage."""
    # Access the Python version directly via the py_func attribute
    py_func = _compute_ccf_rectangle.py_func

    # Create test data
    lags = np.array([0.0, 1.0, 2.0, 3.0])
    t_x = np.array([0.0, 1.0, 2.0, 3.0, 4.0])
    x = random(5)
    t_y = np.array([0.0, 1.0, 2.0, 3.0, 4.0])
    y = random(5)
    bin_width = 0.5

    # Call the Python function directly
    c, b = py_func(lags, t_x, x, t_y, y, bin_width)

    # Simple assertion to ensure the function runs successfully
    assert len(c) == len(lags)
    assert len(b) == len(lags)


def test_compute_ccf_gaussian_py_func() -> None:
    """Test the Python version of _compute_ccf_gaussian for coverage."""
    # Access the Python version directly via the py_func attribute
    py_func = _compute_ccf_gaussian.py_func

    # Create test data
    lags = np.array([0.0, 1.0, 2.0, 3.0])
    t_x = np.array([0.0, 1.0, 2.0, 3.0, 4.0])
    x = random(5)
    t_y = np.array([0.0, 1.0, 2.0, 3.0, 4.0])
    y = random(5)
    bin_width = 0.5

    # Call the Python function directly
    c, b = py_func(lags, t_x, x, t_y, y, bin_width)

    # Simple assertion to ensure the function runs successfully
    assert len(c) == len(lags)
    assert len(b) == len(lags)


def test_rectangle_vs_gaussian_consistency() -> None:
    """Test that rectangle and gaussian methods produce similar results for simple cases."""
    # Create test data with small differences
    lags = np.array([0.0, 1.0, 2.0])
    t_x = np.array([0.0, 1.0, 2.0, 3.0])
    x = np.array([1.0, 2.0, 1.0, 0.0])
    t_y = np.array([0.0, 1.0, 2.0, 3.0])
    y = np.array([0.0, 1.0, 2.0, 1.0])

    # Get Python implementations for both methods
    rect_py = _compute_ccf_rectangle.py_func
    gauss_py = _compute_ccf_gaussian.py_func

    # Calculate with both methods using narrow bins
    c_rect, _ = rect_py(lags, t_x, x, t_y, y, bin_width=0.1)
    c_gauss, _ = gauss_py(lags, t_x, x, t_y, y, bin_width=0.1)

    # With narrow bins, methods should be similar for simple cases
    # (not exactly equal due to different weighting)
    assert_array_almost_equal(c_rect, c_gauss, decimal=1)



"""This module contains the different solvers that are available for Pastas.

All solvers inherit from the BaseSolver class, which contains general method for
selecting the correct time series to misfit and options to weight the residuals or
noise series.

To solve a model the following syntax can be used:

>>> ml.solve(solver=ps.LeastSquares())
"""

import importlib
from collections.abc import Callable
from logging import getLogger
from typing import Literal

import numpy as np
from pandas import DataFrame, Series
from scipy.linalg import LinAlgError, get_lapack_funcs, svd
from scipy.optimize import Bounds, least_squares

from pastas.objective_functions import GaussianLikelihood
from pastas.typing import ArrayLike, CallBack, Model

logger = getLogger(__name__)


class BaseSolver:
    _name = "BaseSolver"
    __doc__ = """All solver instances inherit from the BaseSolver class.

    Attributes
    ----------
    pcov: pandas.DataFrame
        Pandas DataFrame with the correlation between the optimized parameters.
    pcor: pandas.DataFrame
        Based on pcov, cannot be parsed.
        Pandas DataFrame with the correlation between the optimized parameters.
    nfev: int
        Number of times the model is called during optimization.
    result: object
        The object returned by the minimization method that is used. It depends
        on the solver what is actually returned.

    """

    def __init__(
        self,
        pcov: DataFrame | None = None,
        nfev: int | None = None,
        obj_func: Callable | None = None,
        **kwargs,
    ) -> None:
        self.ml = None
        self.pcov = pcov  # Covariances of the parameters
        if pcov is None:
            self.pcor = None  # Correlation between parameters
        else:
            self.pcor = self._get_correlations(pcov)
        self.nfev = nfev  # number of function evaluations
        self.obj_func = obj_func
        self.result = None  # Object returned by the optimization method
        if kwargs:
            logger.warning(
                "kwargs to the solver instance are ignored, please provide the"
                "kwargs to the model.solve method."
            )

    def set_model(self, ml: Model):
        """Method to set the Pastas Model instance.

        Parameters
        ----------
        ml: pastas.Model instance

        """
        if self.ml is not None:
            raise UserWarning(
                "This solver instance is already used by another model. Please create "
                "a separate solver instance for each Pastas Model."
            )
        self.ml = ml

    def misfit(
        self,
        p: ArrayLike,
        noise: bool,
        weights: Series | None = None,
        callback: CallBack | None = None,
        returnseparate: bool = False,
    ) -> ArrayLike | tuple[ArrayLike, ArrayLike, ArrayLike]:
        """This method is called by all solvers to obtain a series that are
        minimized in the optimization process. It handles the application of
        the weights, a noisemodel and other optimization options.

        Parameters
        ----------
        p: array_like
            array_like object with the values as floats representing the
            model parameters.
        noise: Boolean
        weights: pandas.Series, optional
            pandas Series by which the residual or noise series are
            multiplied. Typically values between 0 and 1.
        callback: ufunc, optional
            function that is called after each iteration. the parameters are
            provided to the func. E.g. "callback(parameters)"
        returnseparate: bool, optional
            return residuals, noise, noiseweights

        Returns
        -------
        rv: array_like
            residuals array (if noise=False) or noise array (if noise=True)
        """
        # Get the residuals or the noise
        if noise:
            rv = self.ml.noise(p) * self.ml.noise_weights(p)

        else:
            rv = self.ml.residuals(p)

        # Determine if weights need to be applied
        if weights is not None:
            weights = weights.reindex(rv.index)
            weights.fillna(1.0, inplace=True)
            rv = rv.multiply(weights)

        if callback:
            callback(p)

        if returnseparate:
            return (
                self.ml.residuals(p).values,
                self.ml.noise(p).values,
                self.ml.noise_weights(p).values,
            )

        return rv.values

    def prediction_interval(
        self, n: int = 1000, alpha: float = 0.05, max_iter: int = 10, **kwargs
    ) -> DataFrame:
        """Method to calculate the prediction interval for the simulation.

        Returns
        -------
        data : Pandas.DataFrame
            DataFrame of length number of observations and two columns labeled
            0.025 and 0.975 (numerical values) containing the 2.5% and 97.5%
            prediction interval (for alpha=0.05)
        **kwargs
            Additional keyword arguments are passed to the `ml.simulate()` method.
            For example, `tmin` and `tmax` can be passed as keyword arguments to compute
            the prediction interval for a specific period.

        Notes
        -----
        Add residuals assuming a Normal distribution with standard deviation
        equal to the standard deviation of the residuals.
        """

        sigr = self.ml.residuals().std()

        data = self._get_realizations(
            func=self.ml.simulate, n=n, name=None, max_iter=max_iter, **kwargs
        )
        data = data + sigr * np.random.randn(data.shape[0], data.shape[1])

        q = [alpha / 2, 1 - alpha / 2]
        rv = data.quantile(q, axis=1).transpose()
        return rv

    def ci_simulation(
        self, n: int = 1000, alpha: float = 0.05, max_iter: int = 10, **kwargs
    ) -> DataFrame:
        """Method to calculate the confidence interval for the simulation.

        Returns
        -------
        data : Pandas.DataFrame
            DataFrame of length number of observations and two columns labeled
            0.025 and 0.975 (numerical values) containing the 2.5% and 97.5%
            interval (for alpha=0.05)
        **kwargs
            Additional keyword arguments are passed to the `ml.simulate()` method.
            For example, `tmin` and `tmax` can be passed as keyword arguments to compute
            the confidence interval for a specific period.

        Notes
        -----
        The confidence interval shows the uncertainty in the simulation due
        to parameter uncertainty. In other words, there is a 95% probability
        that the true best-fit line for the observed data lies within the
        95% confidence interval.
        """
        return self._get_confidence_interval(
            func=self.ml.simulate, n=n, alpha=alpha, max_iter=max_iter, **kwargs
        )

    def ci_block_response(
        self,
        name: str,
        n: int = 1000,
        alpha: float = 0.05,
        max_iter: int = 10,
        **kwargs,
    ) -> DataFrame:
        """Method to calculate the confidence interval for the block response.

        Returns
        -------
        data : Pandas.DataFrame
            DataFrame of length number of observations and two columns labeled
            0.025 and 0.975 (numerical values) containing the 2.5% and 97.5%
            interval (for alpha=0.05)
        **kwargs
            Additional keyword arguments are passed to the `ml.get_block_response()`
            method.

        Notes
        -----
        The confidence interval shows the uncertainty in the simulation due
        to parameter uncertainty. In other words, there is a 95% probability
        that the true best-fit line for the observed data lies within the
        95% confidence interval.
        """
        dt = self.ml.get_block_response(name=name).index.values
        return self._get_confidence_interval(
            func=self.ml.get_block_response,
            n=n,
            alpha=alpha,
            name=name,
            max_iter=max_iter,
            dt=dt,
            **kwargs,
        )

    def ci_step_response(
        self,
        name: str,
        n: int = 1000,
        alpha: float = 0.05,
        max_iter: int = 10,
        **kwargs,
    ) -> DataFrame:
        """Method to calculate the confidence interval for the step response.

        Returns
        -------
        data : Pandas.DataFrame
            DataFrame of length number of observations and two columns labeled
            0.025 and 0.975 (numerical values) containing the 2.5% and 97.5%
            interval (for alpha=0.05)
        **kwargs
            Additional keyword arguments are passed to the `ml.get_step_response()`
            method.

        Notes
        -----
        The confidence interval shows the uncertainty in the simulation due
        to parameter uncertainty. In other words, there is a 95% probability
        that the true best-fit line for the observed data lies within the
        95% confidence interval.
        """
        dt = self.ml.get_block_response(name=name).index.values
        return self._get_confidence_interval(
            func=self.ml.get_step_response,
            n=n,
            alpha=alpha,
            name=name,
            max_iter=max_iter,
            dt=dt,
            **kwargs,
        )

    def ci_contribution(
        self,
        name: str,
        n: int = 1000,
        alpha: float = 0.05,
        max_iter: int = 10,
        **kwargs,
    ) -> DataFrame:
        """Method to calculate the confidence interval for the contribution.

        Returns
        -------
        data : Pandas.DataFrame
            DataFrame of length number of observations and two columns labeled
            0.025 and 0.975 (numerical values) containing the 2.5% and 97.5%
            interval (for alpha=0.05).
        **kwargs
            Additional keyword arguments are passed to the `ml.get_contribution()`
            method. For example, `tmin` and `tmax` can be passed as keyword arguments to
            compute the confidence interval of a contribution for a specific period.

        Notes
        -----
        The confidence interval shows the uncertainty in the simulation due
        to parameter uncertainty. In other words, there is a 95% probability
        that the true best-fit line for the observed data lies within the
        95% confidence interval.
        """
        return self._get_confidence_interval(
            func=self.ml.get_contribution,
            n=n,
            alpha=alpha,
            name=name,
            max_iter=max_iter,
            **kwargs,
        )

    def get_parameter_sample(
        self, name: str | None = None, n: int = None, max_iter: int = 10
    ) -> ArrayLike:
        """Method to obtain a parameter sets for monte carlo analyses.

        Parameters
        ----------
        name: str, optional
            Name of the stressmodel or model component to obtain the
            parameters for.
        n: int, optional
            Number of random samples drawn from the bivariate normal
            distribution.
        max_iter : int, optional
            maximum number of iterations for truncated multivariate
            sampling, default is 10. Increase this value if number of
            accepted parameter samples is lower than n.

        Returns
        -------
        array_like
            array with N parameter samples.
        """
        p = self.ml.get_parameters(name=name)
        pcov = self._get_covariance_matrix(name=name)

        if name is None:
            parameters = self.ml.parameters
        else:
            parameters = self.ml.parameters.loc[self.ml.parameters.name == name]

        pmin = parameters.pmin.fillna(-np.inf).values
        pmax = parameters.pmax.fillna(np.inf).values

        if n is None:
            # only use parameters that are varied.
            n = int(10 ** parameters.vary.sum())

        samples = np.zeros((0, p.size))

        # Start truncated multivariate sampling
        it = 0
        while samples.shape[0] < n:
            s = np.random.multivariate_normal(p, pcov, size=(n,), check_valid="ignore")
            accept = s[
                (np.min(s - pmin, axis=1) >= 0) & (np.max(s - pmax, axis=1) <= 0)
            ]
            samples = np.concatenate((samples, accept), axis=0)

            # Make sure there's no endless while loop
            if it > max_iter:
                break
            else:
                it += 1

        if samples.shape[0] < n:
            suggestion = "You could try increasing 'max_iter'."
            if samples.shape[0] == 0:
                raise Exception(
                    "No parameter samples were found within %s runs. " % max_iter
                    + suggestion
                )
            else:
                logger.warning(
                    "Parameter sample size is smaller than n: %s/%s. " % (max_iter, n)
                    + suggestion
                )
        return samples[:n, :]

    def _get_realizations(
        self,
        func: Callable,
        n: int | None = None,
        name: str | None = None,
        max_iter: int = 10,
        **kwargs,
    ) -> DataFrame:
        """Internal method to obtain n number of parameter realizations."""
        if name:
            kwargs["name"] = name

        parameter_sample = self.get_parameter_sample(n=n, name=name, max_iter=max_iter)
        data = {}

        for i, p in enumerate(parameter_sample):
            data[i] = func(p=p, **kwargs)

        return DataFrame.from_dict(data, orient="columns", dtype=float)

    def _get_confidence_interval(
        self,
        func: Callable,
        n: int | None = None,
        name: str | None = None,
        max_iter: int = 10,
        alpha: float = 0.05,
        **kwargs,
    ) -> DataFrame:
        """Internal method to obtain a confidence interval."""
        q = [alpha / 2, 1 - alpha / 2]
        data = self._get_realizations(
            func=func, n=n, name=name, max_iter=max_iter, **kwargs
        )
        return data.quantile(q=q, axis=1).transpose()

    def _get_covariance_matrix(self, name: str | None = None) -> DataFrame:
        """Internal method to obtain the covariance matrix from the model.

        Parameters
        ----------
        name: str, optional
            Name of the stressmodel or model component to obtain the
            parameters for.

        Returns
        -------
        pcov: pandas.DataFrame
            Pandas DataFrame with the covariances for the parameters.
        """
        if name:
            index = self.ml.parameters.loc[
                self.ml.parameters.loc[:, "name"] == name
            ].index
        else:
            index = self.ml.parameters.index

        pcov = self.pcov.reindex(index=index, columns=index).fillna(0)

        return pcov

    @staticmethod
    def _get_correlations(pcov: DataFrame) -> DataFrame:
        """Internal method to obtain the parameter correlations from the
        covariance matrix.

        Parameters
        ----------
        pcov: pandas.DataFrame
            n x n Pandas DataFrame with the covariances.

        Returns
        -------
        pcor: pandas.DataFrame
            n x n Pandas DataFrame with the correlations.
        """
        index = pcov.index
        pcov = pcov.to_numpy()
        v = np.sqrt(np.diag(pcov))
        with np.errstate(divide="ignore", invalid="ignore"):
            corr = pcov / np.outer(v, v)
        corr[pcov == 0] = 0
        pcor = DataFrame(data=corr, index=index, columns=index)
        return pcor

    def to_dict(self) -> dict:
        data = {
            "class": self._name,
            "pcov": self.pcov,
            "nfev": self.nfev,
            "obj_func": self.obj_func,
        }
        return data


class LeastSquares(BaseSolver):
    """Solver based on Scipy's least_squares method :cite:p:`virtanen_scipy_2020`.

    Notes
    -----
    This class is the default solve method called by the pastas Model solve
    method. All kwargs provided to the Model.solve() method are forwarded to the
    solver. From there, they are forwarded to Scipy least_squares solver.

    Examples
    --------

    >>> ml.solve(solver=ps.LeastSquares())

    References
    ----------
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.least_squares.html
    """

    _name = "LeastSquares"

    def __init__(
        self,
        pcov: DataFrame | None = None,
        nfev: int | None = None,
        **kwargs,
    ) -> None:
        BaseSolver.__init__(self, pcov=pcov, nfev=nfev, **kwargs)

    def solve(
        self,
        noise: bool = True,
        weights: Series | None = None,
        callback: CallBack | None = None,
        **kwargs,
    ) -> tuple[bool, ArrayLike, ArrayLike]:
        self.vary = self.ml.parameters.vary.values.astype(bool)
        self.initial = self.ml.parameters.initial.values.copy()
        parameters = self.ml.parameters.loc[self.vary]

        # Set the boundaries
        method = kwargs.pop("method") if "method" in kwargs else "trf"
        if method == "lm":
            logger.info(
                "Method 'lm' does not support boundaries. Ignoring Pastas'"
                "`pmin` and `pmax` parameter bounds and setting them to `nan`."
            )
            bounds = Bounds(
                lb=np.full(len(parameters), -np.inf),
                ub=np.full(len(parameters), np.inf),
                keep_feasible=True,
            )
            # set to nan because that's what is used by the solver
            self.ml.parameters.loc[self.vary, "pmin"] = np.nan
            self.ml.parameters.loc[self.vary, "pmax"] = np.nan
        else:
            bounds = Bounds(
                lb=np.where(parameters.pmin.isnull(), -np.inf, parameters.pmin),
                ub=np.where(parameters.pmax.isnull(), np.inf, parameters.pmax),
                keep_feasible=True,
            )

        self.result = least_squares(
            self.objfunction,
            bounds=bounds,
            x0=parameters.initial.values,
            args=(noise, weights, callback),
            method=method,
            **kwargs,
        )

        self.pcov = DataFrame(
            LeastSquares.get_covariances(
                self.result.jac, self.result.cost, method=method, absolute_sigma=False
            ),
            index=parameters.index,
            columns=parameters.index,
        )
        self.pcor = self._get_correlations(self.pcov)
        self.nfev = self.result.nfev
        self.obj_func = self.result.cost

        # Prepare return values
        success = self.result.success
        optimal = self.initial
        optimal[self.vary] = self.result.x
        stderr = np.zeros(len(optimal)) * np.nan
        stderr[self.vary] = np.sqrt(np.diag(self.pcov))

        return success, optimal, stderr

    def objfunction(
        self, p: ArrayLike, noise: bool, weights: Series, callback: CallBack
    ) -> ArrayLike:
        par = self.initial
        par[self.vary] = p
        return self.misfit(p=par, noise=noise, weights=weights, callback=callback)

    @staticmethod
    def get_covariances(
        jacobian: ArrayLike,
        cost: float,
        method: Literal["trf", "dogbox", "lm"] = "trf",
        absolute_sigma: bool = False,
    ) -> ArrayLike:
        """
        Method to get the covariance matrix from the jacobian.

        Parameters
        ----------
        jacobian : ArrayLike
            The jacobian matrix with dimensions nobs, npar.
        cost : float
            The cost value of the scipy.optimize.OptimizeResult which is half
            the sum of squares. That's why the cost is multiplied by a factor
            of two internally to get the sum of squares.
        method : Literal["trf", "dogbox", "lm"], optional
            Algorithm with which the minimization is performed. Default is "trf".
        absolute_sigma : bool, optional
            If True, `sigma` is used in an absolute sense and the estimated
            parameter covariance `pcov` reflects these absolute values. If
            False (default), only the relative magnitudes of the `sigma` values
            matter. The returned parameter covariance matrix `pcov` is based on
            scaling `sigma` by a constant factor. This constant is set by
            demanding that the reduced `chisq` for the optimal parameters
            `popt` when using the *scaled* `sigma` equals unity. In other
            words, `sigma` is scaled to match the sample variance of the
            residuals after the fit. Default is False.
            Mathematically, ``pcov(absolute_sigma=False) =
            pcov(absolute_sigma=True) * chisq(popt)/(M-N)``

        Returns
        -------
        pcov: array_like
            numpy array with the covariance matrix.

        Notes
        -----
        This method is copied from Scipy:
        https://github.com/scipy/scipy/blob/92d2a8592782ee19a1161d0bf3fc2241ba78bb63/scipy/optimize/_minpack_py.py
        Please refer to the SciPy optimization module::
        https://docs.scipy.org/doc/scipy/reference/optimize.html
        """

        nobs, npar = jacobian.shape
        cost = 2 * cost  # res.cost is half sum of squares!
        s_sq = cost / (nobs - npar)  # variance of the residuals

        if method == "lm":
            # https://github.com/scipy/scipy/blob/92d2a8592782ee19a1161d0bf3fc2241ba78bb63/scipy/optimize/_minpack_py.py#L480C9-L499C38
            # fjac A permutation of the R matrix of a QR factorization of the
            # final approximate Jacobian matrix.
            _, fjac = np.linalg.qr(jacobian)
            # leastsq expects the fjacobian to be in fortran order (npar, nobs)
            # that why it is transposed in the original code

            ipvt = np.arange(1, npar + 1, dtype=int)
            n = len(ipvt)
            r = np.triu(fjac[:n, :])

            # old method deprecated in scipy 1.10.0 since
            # the explicit dot product was not necessary and sometimes
            # the result was not symmetric positive definite.
            # See https://github.com/scipy/scipy/issues/4555.
            # old method
            # perm = np.take(np.eye(n), ipvt - 1, 0)
            # R = np.dot(r, perm)
            # cov_x = np.linalg.inv(np.dot(np.transpose(R), R))

            # new method:
            perm = ipvt - 1
            inv_triu = get_lapack_funcs("trtri", (r,))
            try:
                # inverse of permuted matrix is a permutation of matrix inverse
                invR, trtri_info = inv_triu(r)  # default: upper, non-unit diag
                if trtri_info != 0:  # explicit comparison for readability
                    logger.warning(
                        f"LinAlgError in trtri. LAPACK trtri returned info: {trtri_info}"
                    )
                    raise LinAlgError
                invR[perm] = invR.copy()
                pcov = invR @ invR.T  # cov_x in the original code
            except (LinAlgError, ValueError):
                pcov = None
        else:
            # https://github.com/scipy/scipy/blob/92d2a8592782ee19a1161d0bf3fc2241ba78bb63/scipy/optimize/_minpack_py.py#L1029-L1055
            # Do Moore-Penrose inverse discarding zero singular values.
            _, s, VT = svd(jacobian, full_matrices=False)
            threshold = np.finfo(float).eps * max(jacobian.shape) * s[0]
            s = s[s > threshold]
            VT = VT[: s.size]
            pcov = np.dot(VT.T / s**2, VT)

        if pcov is None or np.isnan(pcov).any():
            # indeterminate covariance
            pcov = np.full(shape=(npar, npar), fill_value=np.inf, dtype=float)
            logger.warning(
                "Covariance of the parameters could not be estimated. "
                "The covariance of the parameters is set to infinity."
            )
        elif not absolute_sigma:
            if nobs > npar:
                pcov = pcov * s_sq
            else:
                pcov = np.full(shape=(npar, npar), fill_value=np.inf, dtype=float)
                logger.warning(
                    "Covariance of the parameters could not be estimated. "
                    "The covariance of the parameters is set to infinity."
                )

        return pcov


class LmfitSolve(BaseSolver):
    """Solving the model using the LmFit :cite:p:`newville_lmfitlmfit-py_2019`.

        This is basically a wrapper around the scipy solvers, adding some cool
        functionality for boundary conditions.

    Notes
    -----
    https://github.com/lmfit/lmfit-py/
    """

    _name = "LmfitSolve"

    def __init__(
        self,
        pcov: DataFrame | None = None,
        nfev: int | None = None,
        **kwargs,
    ) -> None:
        try:
            global lmfit
            import lmfit as lmfit  # Import Lmfit here, so it is no dependency
        except ImportError:
            msg = "lmfit not installed. Please install lmfit first."
            raise ImportError(msg) from None
        BaseSolver.__init__(self, pcov=pcov, nfev=nfev, **kwargs)

    def solve(
        self,
        noise: bool = True,
        weights: Series | None = None,
        callback: CallBack | None = None,
        method: str | None = "leastsq",
        **kwargs,
    ) -> tuple[bool, ArrayLike, ArrayLike]:
        # Deal with the parameters
        parameters = lmfit.Parameters()
        p = self.ml.parameters.loc[:, ["initial", "pmin", "pmax", "vary"]]
        for k in p.index:
            pp = np.where(p.loc[k].isnull(), None, p.loc[k])
            parameters.add(k, value=pp[0], min=pp[1], max=pp[2], vary=pp[3])

        # Create the Minimizer object and minimize
        self.mini = lmfit.Minimizer(
            userfcn=self.objfunction,
            calc_covar=True,
            fcn_args=(noise, weights, callback),
            params=parameters,
            **kwargs,
        )
        self.result = self.mini.minimize(method=method)

        # Set all parameter attributes
        pcov = None
        if hasattr(self.result, "covar"):
            if self.result.covar is not None:
                pcov = self.result.covar

        names = self.result.var_names
        self.pcov = DataFrame(pcov, index=names, columns=names, dtype=float)
        self.pcor = self._get_correlations(self.pcov)

        # Set all optimization attributes
        self.nfev = self.result.nfev
        self.obj_func = self.result.chisqr

        if hasattr(self.result, "success"):
            success = self.result.success
        else:
            success = True
        optimal = np.array([p.value for p in self.result.params.values()])
        stderr = np.array([p.stderr for p in self.result.params.values()])

        idx = None
        if "is_weighted" in kwargs:
            if not kwargs["is_weighted"]:
                idx = -1

        return success, optimal[:idx], stderr[:idx]

    def objfunction(
        self, parameters: DataFrame, noise: bool, weights: Series, callback: CallBack
    ) -> ArrayLike:
        p = np.array([p.value for p in parameters.values()])
        return self.misfit(p=p, noise=noise, weights=weights, callback=callback)


class EmceeSolve(BaseSolver):
    """Solver based on MCMC approach in emcee :cite:p:`foreman-mackey_emcee_2013`.

    Parameters
    ----------
    objective_function: func, optional
        An objective function to be minimized. If not provided, the
        GaussianLikelihood is used. See the pastas.objective_functions module for
        more information.
    nwalkers: int, optional
        Number of walkers to use. Default is 20.
    backend: emcee.backend, optional
        One of the Backends from Emcee used to store MCMC results. See Emcee
        for more information.
    moves: emcee.moves, optional
        The moves argument determines how the next step for a walker is chosen in
        the MCMC approach. One of the Moves classes from Emcee has to be provided.
        See Emcee documentation for more information.
    parallel: bool, optional
        Run the sampler in parallel or not.
    progress_bar: bool, optional
        Show the progress bar or not. Requires the `tqdm` package to be installed.
    **kwargs, optional
        All other keyword arguments are passed on to the BaseSolver class.

    Notes
    -----
    The EmceeSolve solver uses the emcee package to perform a Markov Chain Monte Carlo
    (MCMC) approach to find the optimal parameter values. The solver can be used as
    follows:

    >>> solver = ps.EmceeSolve(
    ...     nwalkers=20,
    ...     progress_bar=True,
    ...     )
    >>> ml.solve(solver=solver)

    The arguments provided are mostly passed on to the `emcee.EnsembleSampler`
    and determine how that instance is created. Arguments you want to pass on to
    `run_mcmc` (and indirectly the `sample` method), can be passed on to
    `Model.solve`, like:

    >>> ml.solve(solver=ps.EmceeSolve(), thin_by=2)

    Examples
    --------

    >>> ml.solve(solver=ps.EmceeSolve(), steps=5000)

    To obtain the MCMC chains, use:

    >>> ml.solver.sampler.get_chain(flat=True, discard=3000)

    References
    ----------
    https://emcee.readthedocs.io/en/stable/

    See Also
    --------
    emcee.EnsembleSampler
    emcee.moves
    emcee.backend
    pastas.objective_functions

    """

    _name = "EmceeSolve"

    def __init__(
        self,
        objective_function=None,
        nwalkers: int = 20,
        backend=None,
        moves=None,
        parallel: bool = False,
        progress_bar: bool = True,
        **kwargs,
    ) -> None:
        # Check if emcee is installed, if not, return error
        try:
            global emcee
            import emcee as emcee  # Import emcee here, so it is no dependency
        except ImportError:
            msg = "emcee not installed. Please install emcee first."
            raise ImportError(msg) from None

        BaseSolver.__init__(self, pcov=None, nfev=None, **kwargs)

        # Set Attributes
        self.obj_func = np.nan
        self.nfev = np.nan

        # Set sampler properties
        self.sampler = None
        self.parallel = parallel
        self.backend = backend
        self.moves = moves
        self.progress_bar = progress_bar
        self.nwalkers = nwalkers
        self.priors = None

        # Set objective function
        if objective_function is None:
            objective_function = GaussianLikelihood()
        self.objective_function = objective_function
        self.parameters = self.objective_function.get_init_parameters("ln")

    def solve(
        self,
        noise: bool = False,
        weights: Series | None = None,
        steps: int = 5000,
        callback: CallBack | None = None,
        **kwargs,
    ) -> tuple[bool, ArrayLike, ArrayLike]:
        # Store initial parameters
        self.initial = np.append(
            self.ml.parameters.initial.values, self.parameters.initial.values
        )
        self.vary = np.append(
            self.ml.parameters.vary.values, self.parameters.vary.values
        )

        # Set lower and upper bounds
        lb = np.append(
            self.ml.parameters[self.ml.parameters.vary].pmin.values,
            self.parameters[self.parameters.vary].pmin.values,
        )
        ub = np.append(
            self.ml.parameters[self.ml.parameters.vary].pmax.values,
            self.parameters[self.parameters.vary].pmax.values,
        )
        self.bounds = np.vstack([lb, ub]).T

        # Set priors
        self._set_priors()

        # Set initial positions of the walkers
        pinit = np.append(
            self.ml.parameters[self.ml.parameters.vary].initial.values,
            self.parameters[self.parameters.vary].initial.values,
        )
        ndim = pinit.size

        pinit = pinit + np.abs(pinit) * 1e-2 * np.random.randn(self.nwalkers, ndim)

        # Create sampler and run mcmc
        if self.parallel:
            logger.info("Going into the parallel universe")

            from multiprocessing import Pool

            with Pool() as pool:
                self.sampler = emcee.EnsembleSampler(
                    nwalkers=self.nwalkers,
                    ndim=ndim,
                    log_prob_fn=self.log_probability,
                    moves=self.moves,
                    backend=self.backend,
                    pool=pool,
                    args=(noise, weights, callback),
                )

                self.sampler.run_mcmc(
                    pinit, steps, progress=self.progress_bar, **kwargs
                )
        else:
            self.sampler = emcee.EnsembleSampler(
                nwalkers=self.nwalkers,
                ndim=ndim,
                log_prob_fn=self.log_probability,
                moves=self.moves,
                backend=self.backend,
                pool=None,
                args=(noise, weights, callback),
            )

            self.sampler.run_mcmc(pinit, steps, progress=self.progress_bar, **kwargs)

        # Get optimal values
        optimal = self.initial.copy()
        chains = self.sampler.get_chain(discard=0, flat=True, thin=1)
        optimal[self.vary] = chains[self.sampler.get_log_prob().argmax()]

        # Set the optimal values for the objective function parameters
        self.parameters.loc[:, "optimal"] = optimal[-self.objective_function.nparam :]

        # Don't estimate stderr for now
        optimal = optimal[: -self.objective_function.nparam]
        stderr = np.zeros(len(optimal)) * np.nan

        success = True
        return success, optimal, stderr

    def log_probability(
        self,
        p: ArrayLike,
        noise: bool | None = False,
        weights: Series | None = None,
        callback: CallBack | None = None,
    ) -> float:
        """Full log-probability called by Emcee.

        Parameters
        ----------
        p: numpy.Array
            Numpy array with the parameters.
        noise: bool, optional
            If True, the noise model is applied to the residuals.
        weights: pandas.Series, optional
            Series with weights for the residuals.
        callback: callable, optional
            Callback function that will be called after each iteration of the solver.

        Returns
        -------
        log_probability: float

        """
        lp = self.log_prior(p)

        # This will occur if the parameters are outside the boundaries
        if not np.isfinite(lp):
            return -np.inf
        else:
            return lp + self.log_likelihood(
                p, noise=noise, weights=weights, callback=callback
            )

    def log_likelihood(
        self,
        p: ArrayLike,
        noise: bool,
        weights: Series | None = None,
        callback: CallBack | None = None,
    ) -> float:
        """Log-likelihood function.

        Parameters
        ----------
        p: numpy.Array
            Numpy array with the parameters.
        noise: bool

        weights
        callback

        Returns
        -------
        lnlike: float
            The log-likelihood for the parameters.

        Notes
        -----
        This method is always called by emcee.

        """
        par = self.initial

        # Set the parameters that are varied from the model and objective function
        par[self.vary] = p

        rv = self.misfit(
            p=par[: -self.objective_function.nparam],
            noise=noise,
            weights=weights,
            callback=callback,
        )

        lnlike = self.objective_function.compute(
            rv, par[-self.objective_function.nparam :]
        )

        return lnlike

    def log_prior(self, p: ArrayLike) -> float:
        """Probability of parameter set given the priors.

        Parameters
        ----------
        p: numpy.Array
            Numpy array with the parameters

        Returns
        -------
        lp: float
            Probability of parameter set given the priors

        Notes
        -----
        Two cases exist:

        - If any of the parameters touch the boundary, -np.inf is returned. This
          basically tells the algorithm that the parameter set is very unlikely.
        - Otherwise, the probability of each parameter given its prior is computed.

        """
        # Check if parameters are within the boundaries
        if np.any(p < self.bounds[:, 0]) or np.any(p > self.bounds[:, 1]):
            lp = -np.inf
        # If not, compute the probability of each parameter given its prior
        else:
            lp = 0.0
            for param, prior in zip(p, self.priors):
                lp += prior.logpdf(param)
        return lp

    def _set_priors(self) -> None:
        """Set the priors for the parameters."""
        self.priors = []

        # Set the priors for the parameters that are varied from the model
        for _, (loc, pmin, pmax, scale, dist) in self.ml.parameters.loc[
            self.ml.parameters.vary, ["initial", "pmin", "pmax", "stderr", "dist"]
        ].iterrows():
            self.priors.append(self._get_prior(dist, loc, scale, pmin, pmax))

        # Set the priors for the parameters that are varied from the objective function
        for _, (loc, pmin, pmax, scale, dist) in self.parameters.loc[
            self.parameters.vary, ["initial", "pmin", "pmax", "stderr", "dist"]
        ].iterrows():
            self.priors.append(self._get_prior(dist, loc, scale, pmin, pmax))

    def _get_prior(self, dist: str, loc: float, scale: float, pmin: float, pmax: float):
        """Set the prior for a parameter.

        Parameters
        ----------
        dist: str
            Name of the distribution. Must be a scipy.stats distribution.
        loc: float
            Location parameter. For example, the mean for a normal distribution.
        scale: float
            Scale parameter. For example, the standard deviation for a normal distribution.

        Returns
        -------
        dist: scipy.stats distribution

        """
        # Import the distribution
        mod = importlib.import_module("scipy.stats")
        # Return the distribution
        if dist == "uniform":
            loc = pmin
            scale = pmax - pmin

        if np.isnan(loc) or np.isnan(scale):
            msg = "Location and/or scale parameter is NaN."
            logger.error(msg)
            raise ValueError(msg)

        return getattr(mod, dist)(loc=loc, scale=scale)

    def set_parameter(
        self,
        name: str,
        initial: float | None = None,
        vary: bool | None = None,
        pmin: float | None = None,
        pmax: float | None = None,
        optimal: float | None = None,
        dist: str | None = None,
    ) -> None:
        """Method to change the parameter properties.

        Parameters
        ----------
        name: str
            name of the parameter to update. This has to be a single variable.
        initial: float, optional
            parameters value to use as initial estimate.
        vary: bool, optional
            boolean to vary a parameter (True) or not (False).
        pmin: float, optional
            minimum value for the parameter.
        pmax: float, optional
            maximum value for the parameter.
        optimal: float, optional
            optimal value for the parameter.
        dist: str, optional
            Distribution of the parameters. Must be a scipy.stats distribution.

        Examples
        --------
        >>> s = ps.EmceeSolve()
        >>> s.set_parameter(name="ln_sigma", initial=0.1, vary=True, pmin=0.01, pmax=1)

        Notes
        -----
        It is highly recommended to use this method to set parameter properties.
        Changing the parameter properties directly in the parameter `DataFrame` may
        not work as expected.

        """
        # Check if the parameter is present in the solver
        if name not in self.parameters.index:
            msg = "parameter %s is not present in the solver."
            self.logger.error(msg, name)
            raise KeyError(msg % name)

        # Set the initial value
        if initial is not None:
            self.parameters.at[name, "initial"] = float(initial)

        # Set the vary property
        if vary is not None:
            self.parameters.at[name, "vary"] = bool(vary)

        # Set the minimum value
        if pmin is not None:
            self.parameters.at[name, "pmin"] = float(pmin)

        # Set the maximum value
        if pmax is not None:
            self.parameters.at[name, "pmax"] = float(pmax)

        # Set the optimal value
        if optimal is not None:
            self.parameters.at[name, "optimal"] = float(optimal)

        # Set the distribution
        if dist is not None:
            self.parameters.at[name, "dist"] = str(dist)

    def to_dict(self) -> dict:
        """This method is not supported for this solver.

        Raises
        ------
        NotImplementedError

        """
        msg = "The EmceeSolve class does not support to_dict() and cannot be saved."
        raise NotImplementedError(msg)



"""Tests for the noise models in pastas."""

import numpy as np
import pandas as pd
import pytest
from pandas import Series

from pastas.noisemodels import ArmaNoiseModel, ArNoiseModel


@pytest.fixture
def residual_series() -> Series:
    """Create a residual series for testing."""
    # Create daily data for 100 days
    dates = pd.date_range(start="2020-01-01", periods=100, freq="D")
    # Create a residual series with some pattern
    values = np.sin(np.linspace(0, 2 * np.pi * 3, 100)) + np.random.normal(0, 0.1, 100)
    return Series(values, index=dates, name="Residuals")


@pytest.fixture
def irregular_residual_series() -> Series:
    """Create an irregular residual series for testing."""
    # Create 100 dates with irregular spacing
    base_dates = pd.date_range(start="2020-01-01", periods=150, freq="D")
    # Select random 100 dates (sorted)
    selected_indices = np.sort(np.random.choice(150, 100, replace=False))
    dates = base_dates[selected_indices]

    # Create residuals with some pattern
    values = np.sin(np.linspace(0, 2 * np.pi * 3, 100)) + np.random.normal(0, 0.1, 100)
    return Series(values, index=dates, name="Residuals")


class TestArNoiseModel:
    """Test the AR noise model."""

    def test_init(self) -> None:
        """Test initialization of AR noise model."""
        model = ArNoiseModel()
        assert model._name == "ArNoiseModel"
        assert model.nparam == 1
        assert model.norm is True
        assert "noise_alpha" in model.parameters.index

        # Test initialization with norm=False
        model = ArNoiseModel(norm=False)
        assert not model.norm

    def test_set_init_parameters(self) -> None:
        """Test setting initial parameters with and without oseries."""
        # Without oseries
        model = ArNoiseModel()
        model.set_init_parameters()
        assert model.parameters.loc["noise_alpha", "initial"] == 14.0

        # With oseries
        oseries = pd.Series(
            index=pd.date_range(start="2020-01-01", periods=10, freq="3D")
        )
        model = ArNoiseModel()
        model.set_init_parameters(oseries)
        assert model.parameters.loc["noise_alpha", "initial"] == 3.0  # 3D frequency

    def test_simulate(self, residual_series: Series) -> None:
        """Test noise simulation."""
        model = ArNoiseModel()
        alpha = 10.0  # Set alpha parameter

        # Simulate noise
        noise = model.simulate(residual_series, [alpha])

        # Check properties
        assert len(noise) == len(residual_series)
        assert isinstance(noise, Series)
        assert noise.name == "Noise"

        # First value should equal first residual
        assert noise.iloc[0] == residual_series.iloc[0]

        # Calculate expected noise for second value
        dt = (residual_series.index[1] - residual_series.index[0]).days
        expected_noise = (
            residual_series.iloc[1] - np.exp(-dt / alpha) * residual_series.iloc[0]
        )
        assert noise.iloc[1] == pytest.approx(expected_noise)

    def test_weights(self, residual_series: Series) -> None:
        """Test weights calculation."""
        model = ArNoiseModel()
        alpha = 10.0

        # Get weights
        weights = model.weights(residual_series, [alpha])

        # Check basic properties
        assert len(weights) == len(residual_series)
        assert isinstance(weights, Series)
        assert weights.name == "noise_weights"

        # Test with normalization off
        model.norm = False
        weights_no_norm = model.weights(residual_series, [alpha])

        # Should be different from normalized weights
        assert not np.allclose(weights.values, weights_no_norm.values)

    def test_get_correction(self, residual_series: Series) -> None:
        """Test forecast correction."""
        model = ArNoiseModel()
        alpha = 10.0

        # Generate forecast dates
        forecast_dates = pd.date_range(
            start=residual_series.index[-1] + pd.Timedelta(days=1), periods=10, freq="D"
        )

        # Get correction
        correction = model.get_correction(residual_series, [alpha], forecast_dates)

        # Check properties
        assert len(correction) == len(forecast_dates)
        assert correction.name == "correction"

        # Check calculation
        last_residual = residual_series.iloc[-1]
        dt_days = 1  # First forecast is 1 day away
        expected_first_correction = np.exp(-dt_days / alpha) * last_residual
        assert correction.iloc[0] == pytest.approx(expected_first_correction)

        # Correction should decay with time
        assert abs(correction.iloc[-1]) < abs(correction.iloc[0])

    def test_to_dict(self) -> None:
        """Test conversion to dictionary."""
        model = ArNoiseModel(norm=False)
        data = model.to_dict()

        assert data["class"] == "ArNoiseModel"
        assert data["norm"] is False


class TestArmaNoiseModel:
    """Test the ARMA noise model."""

    def test_init(self) -> None:
        """Test initialization of ARMA noise model."""
        model = ArmaNoiseModel()
        assert model._name == "ArmaNoiseModel"
        assert model.nparam == 2
        assert "noise_alpha" in model.parameters.index
        assert "noise_beta" in model.parameters.index

    def test_set_init_parameters(self) -> None:
        """Test setting initial parameters with and without oseries."""
        # Without oseries
        model = ArmaNoiseModel()
        model.set_init_parameters()
        assert model.parameters.loc["noise_alpha", "initial"] == 14.0
        assert model.parameters.loc["noise_beta", "initial"] == 1.0

        # With oseries
        oseries = pd.Series(
            index=pd.date_range(start="2020-01-01", periods=10, freq="7D")
        )
        model = ArmaNoiseModel()
        model.set_init_parameters(oseries)
        assert model.parameters.loc["noise_alpha", "initial"] == 7.0  # 7D frequency
        assert model.parameters.loc["noise_beta", "initial"] == 1.0

    def test_simulate(self, residual_series: Series) -> None:
        """Test noise simulation."""
        model = ArmaNoiseModel()
        params = [10.0, 5.0]  # alpha, beta

        # Simulate noise
        noise = model.simulate(residual_series, params)

        # Check properties
        assert len(noise) == len(residual_series)
        assert isinstance(noise, Series)
        assert noise.name == "Noise"

        # First value should equal first residual
        assert noise.iloc[0] == residual_series.iloc[0]

    def test_calculate_noise_edge_cases(self, residual_series: Series) -> None:
        """Test noise calculation with edge case parameters."""
        model = ArmaNoiseModel()

        # Test with beta = 0 (should use a small value instead)
        params = [10.0, 0.0]  # alpha, beta
        noise_beta_zero = model.simulate(residual_series, params)
        assert np.isfinite(noise_beta_zero).all()

        # Test with negative beta (should handle sign correctly)
        params = [10.0, -5.0]  # alpha, beta
        noise_beta_neg = model.simulate(residual_series, params)
        assert np.isfinite(noise_beta_neg).all()


class TestParameterSetting:
    """Test parameter setting methods."""

    def test_set_parameter_methods(self) -> None:
        """Test parameter setting methods."""
        model = ArNoiseModel()

        # Test setting initial value
        model._set_initial("noise_alpha", 20.0)
        assert model.parameters.loc["noise_alpha", "initial"] == 20.0

        # Test setting min/max
        model._set_pmin("noise_alpha", 5.0)
        model._set_pmax("noise_alpha", 100.0)
        assert model.parameters.loc["noise_alpha", "pmin"] == 5.0
        assert model.parameters.loc["noise_alpha", "pmax"] == 100.0

        # Test setting vary
        model._set_vary("noise_alpha", False)
        assert not model.parameters.loc["noise_alpha", "vary"]

        # Test setting distribution
        model._set_dist("noise_alpha", "normal")
        assert model.parameters.loc["noise_alpha", "dist"] == "normal"


def test_irregular_time_steps(irregular_residual_series: Series) -> None:
    """Test noise models with irregular time steps."""
    ar_model = ArNoiseModel()

    # Set parameters
    ar_params = [10.0]  # alpha

    # Simulate noise
    ar_noise = ar_model.simulate(irregular_residual_series, ar_params)

    # Check properties
    assert len(ar_noise) == len(irregular_residual_series)

    # Calculate weights for AR model
    weights = ar_model.weights(irregular_residual_series, ar_params)
    assert len(weights) == len(irregular_residual_series)
